import numpy as np
import pytest
import torch
import scipy.sparse.linalg

from pymde import quadratic, util
import pymde.testing as testing


def test_pca():
    torch.random.manual_seed(0)
    n = 5
    Y = np.random.randn(n, n).astype(np.float32)
    Y -= Y.mean(axis=0)
    for m in range(1, 5):
        X = quadratic.pca(Y, m)
        testing.assert_allclose(1.0 / n * X.T @ X, np.eye(m))
        U, _, _ = np.linalg.svd(Y)
        X_unscaled = 1.0 / np.sqrt(n) * X
        U = U[:, :m]
        U = util.align(source=U, target=X_unscaled)
        for col in range(m):
            testing.assert_allclose(
                X_unscaled[:, col], U[:, col], up_to_sign=True
            )

    n = 5
    k = 4
    Y = np.random.randn(n, k).astype(np.float32)
    Y -= Y.mean(axis=0)
    for m in range(1, k):
        X = quadratic.pca(Y, m)
        testing.assert_allclose(1.0 / n * X.T @ X, np.eye(m))
        U, _, _ = np.linalg.svd(Y)
        X_unscaled = 1.0 / np.sqrt(n) * X
        U = U[:, :m]
        U = util.align(source=U, target=X_unscaled)
        for col in range(m):
            testing.assert_allclose(
                X_unscaled[:, col], U[:, col], up_to_sign=True
            )
    with pytest.raises(
        ValueError, match=r"Embedding dimension must be at most.*"
    ):
        X = quadratic.pca(Y, k + 1)

    n = 4
    k = 5
    Y = np.random.randn(n, k).astype(np.float32)
    Y -= Y.mean(axis=0)
    for m in range(1, n):
        X = quadratic.pca(Y, m)
        testing.assert_allclose(1.0 / n * X.T @ X, np.eye(m))
        U, _, _ = np.linalg.svd(Y)
        X_unscaled = 1.0 / np.sqrt(n) * X
        U = U[:, :m]
        for col in range(m):
            testing.assert_allclose(
                X_unscaled[:, col], U[:, col], up_to_sign=True
            )
    with pytest.raises(
        ValueError, match=r"Embedding dimension must be at most.*"
    ):
        X = quadratic.pca(Y, n + 1)


def test_spectral():
    util.seed(0)
    n = 12
    m = 3
    L = -np.abs(np.random.randn(n, n).astype(np.float32))
    L += L.T
    np.fill_diagonal(L, 0.0)
    np.fill_diagonal(L, -L.sum(axis=1))
    offdiag = np.triu_indices(n, 1)
    edges = np.column_stack(offdiag)
    weights = -L[offdiag]
    X = quadratic.spectral(n, m, torch.tensor(edges), torch.tensor(weights))
    testing.assert_allclose(1.0 / n * X.T @ X, np.eye(m))
    X *= 1.0 / np.sqrt(n)

    eigenvalues, eigenvectors = scipy.sparse.linalg.eigsh(
        L, k=m + 1, which="SM", return_eigenvectors=True
    )
    eigenvectors = eigenvectors[:, 1:]
    for col in range(m):
        testing.assert_allclose(
            eigenvectors[:, col], X[:, col], up_to_sign=True, rtol=1e-3
        )

    # test cg=True
    X = quadratic.spectral(
        n, m, torch.tensor(edges), torch.tensor(weights), cg=True
    )
    X *= 1.0 / np.sqrt(n)
    for col in range(m):
        testing.assert_allclose(
            eigenvectors[:, col], X[:, col], up_to_sign=True, rtol=1e-3
        )

    # test cg=True with torch
    X = quadratic.spectral(
        n, m, torch.tensor(edges), torch.tensor(weights), cg=True
    )
    X *= 1.0 / np.sqrt(n)
    for col in range(m):
        testing.assert_allclose(
            eigenvectors[:, col], X[:, col], up_to_sign=True, rtol=1e-3
        )



from pymde.preprocess.preprocess import dissimilar_edges
from pymde.preprocess.preprocess import deduplicate_edges, sample_edges
from pymde.preprocess.preprocess import scale

from pymde.preprocess.generic import distances, k_nearest_neighbors

from . import graph
from .graph import Graph

from . import data_matrix



import numpy as np
import pytest
import scipy.sparse as sp
import torch

from pymde import preprocess
import pymde.testing as testing


@testing.cpu_and_cuda
def test_from_edges(device):
    edges = torch.tensor([[0, 1], [1, 2], [2, 3], [3, 0]], device=device)
    graph = preprocess.Graph.from_edges(edges)
    assert graph.n_items == 4
    assert graph.n_edges == edges.shape[0]

    expected_edges = torch.tensor([[0, 1], [0, 3], [1, 2], [2, 3]])
    testing.assert_all_equal(expected_edges, graph.edges)

    expected_distances = torch.ones(graph.n_edges)
    testing.assert_all_equal(expected_distances, graph.distances)


@testing.cpu_and_cuda
def test_adjacency_matrix(device):
    edges = torch.tensor([[0, 1], [1, 2], [2, 3], [3, 0]], device=device)
    graph = preprocess.Graph.from_edges(edges)
    expected = np.array(
        [[0, 1, 0, 1], [1, 0, 1, 0], [0, 1, 0, 1], [1, 0, 1, 0]]
    )
    testing.assert_all_equal(expected, graph.A.todense())


@testing.cpu_and_cuda
def test_neighbors(device):
    edges = torch.tensor([[0, 1], [1, 2], [2, 3], [3, 0]], device=device)
    graph = preprocess.Graph.from_edges(edges)

    neighbors = graph.neighbors(0)
    testing.assert_all_equal(np.array([1, 3]), neighbors)

    neighbors = graph.neighbors(1)
    testing.assert_all_equal(np.array([0, 2]), neighbors)

    neighbors = graph.neighbors(2)
    testing.assert_all_equal(np.array([1, 3]), neighbors)

    neighbors = graph.neighbors(3)
    testing.assert_all_equal(np.array([0, 2]), neighbors)


@testing.cpu_and_cuda
def test_distances(device):
    edges = torch.tensor([[0, 1], [1, 2], [2, 3], [3, 0]], device=device)
    graph = preprocess.Graph.from_edges(edges, torch.arange(edges.shape[0]) + 1)

    distances = graph.neighbor_distances(0)
    testing.assert_all_equal(np.array([1.0, 4.0]), distances)

    distances = graph.neighbor_distances(1)
    testing.assert_all_equal(np.array([1.0, 2.0]), distances)

    distances = graph.neighbor_distances(2)
    testing.assert_all_equal(np.array([2.0, 3.0]), distances)

    distances = graph.neighbor_distances(3)
    testing.assert_all_equal(np.array([4.0, 3.0]), distances)


@testing.cpu_and_cuda
def test_distances_repeated_edges_summed(device):
    edges = torch.tensor(
        [[0, 1], [1, 2], [2, 3], [3, 0], [0, 3]], device=device
    )
    graph = preprocess.Graph.from_edges(edges, torch.arange(edges.shape[0]) + 1)

    # edge (0, 3) repeated, so distance is 4 + 5 = 9
    distances = graph.neighbor_distances(0)
    testing.assert_all_equal(np.array([1.0, 9.0]), distances)

    distances = graph.neighbor_distances(1)
    testing.assert_all_equal(np.array([1.0, 2.0]), distances)

    distances = graph.neighbor_distances(2)
    testing.assert_all_equal(np.array([2.0, 3.0]), distances)

    distances = graph.neighbor_distances(3)
    testing.assert_all_equal(np.array([9.0, 3.0]), distances)


def test_from_matrix():
    A = sp.csr_matrix(
        np.array([[0, 2, 0, 3], [2, 0, 1, 1], [0, 1, 0, 0], [3, 1, 0, 0]])
    )
    graph = preprocess.graph.Graph(A)

    expected_edges = torch.tensor([(0, 1), (0, 3), (1, 2), (1, 3)])
    assert graph.n_items == 4
    assert graph.n_edges == 4

    testing.assert_all_equal(expected_edges, graph.edges)
    testing.assert_all_equal(np.array([2.0, 3.0, 1.0, 1.0]), graph.distances)


@testing.cpu_and_cuda
def test_shortest_paths_unweighted(device):
    edges = torch.tensor([[0, 1], [1, 2], [2, 3], [3, 0]], device=device)
    graph = preprocess.Graph.from_edges(edges)

    shortest_path_graph = preprocess.graph.shortest_paths(graph)

    assert shortest_path_graph.n_items == 4
    assert shortest_path_graph.n_edges == 6

    assert shortest_path_graph[0, 1] == 1.0
    assert shortest_path_graph[1, 2] == 1.0
    assert shortest_path_graph[2, 3] == 1.0
    assert shortest_path_graph[3, 0] == 1.0

    assert shortest_path_graph[0, 2] == 2.0
    assert shortest_path_graph[1, 3] == 2.0


@testing.cpu_and_cuda
def test_shortest_paths_unweighted_max_length(device):
    n = 1000
    edges = torch.tensor(
        [[i, i + 1] for i in range(n - 1)] + [[n - 1, 0]], device=device
    )
    graph = preprocess.Graph.from_edges(edges)

    shortest_path_graph = preprocess.graph.shortest_paths(graph, max_length=2)

    assert shortest_path_graph.n_items == n
    assert shortest_path_graph.n_edges == (n - 1 + (4 + n - 3))

    for i in range(n - 1):
        assert shortest_path_graph[i, i + 1] == 1.0
        assert shortest_path_graph[i, (i + 2) % n] == 2.0
        assert shortest_path_graph[i, (i - 2) % n] == 2.0


@testing.cpu_and_cuda
def test_shortest_paths_weighted_max_length(device):
    n = 1000
    edges = torch.tensor(
        [[i, i + 1] for i in range(n - 1)] + [[n - 1, 0]], device=device
    )
    distances = 2.0 * torch.ones(
        edges.shape[0], dtype=torch.float, device=device
    )
    graph = preprocess.Graph.from_edges(edges, distances)

    shortest_path_graph = preprocess.graph.shortest_paths(graph, max_length=4)

    assert shortest_path_graph.n_items == n
    assert shortest_path_graph.n_edges == (n - 1 + (4 + n - 3))

    for i in range(n - 1):
        assert shortest_path_graph[i, i + 1] == 2.0
        assert shortest_path_graph[i, (i + 2) % n] == 4.0
        assert shortest_path_graph[i, (i - 2) % n] == 4.0


@testing.cpu_and_cuda
def test_shortest_paths_weighted(device):
    edges = torch.tensor([[0, 1], [1, 2], [2, 3], [3, 0]], device=device)
    distances = torch.tensor([1.0, 2.0, 3.0, 3.0], device=device)
    graph = preprocess.Graph.from_edges(edges, distances)

    shortest_path_graph = preprocess.graph.shortest_paths(graph)

    assert shortest_path_graph.n_items == 4
    assert shortest_path_graph.n_edges == 6

    assert shortest_path_graph[0, 1] == 1.0
    assert shortest_path_graph[1, 2] == 2.0
    assert shortest_path_graph[2, 3] == 3.0
    assert shortest_path_graph[3, 0] == 3.0

    # 0 -> 1 -> 2
    assert shortest_path_graph[0, 2] == 3.0

    # 1 -> 0 -> 3
    assert shortest_path_graph[1, 3] == 4.0


@testing.cpu_and_cuda
def test_shortest_paths_retain_fraction(device):
    pytest.skip("This test is flaky.")

    # graph on 10 nodes: 10(9)/2 = 45 edges
    n = 10
    edges = torch.tensor(
        [[i, i + 1] for i in range(n - 1)] + [[n - 1, 0]], device=device
    )
    graph = preprocess.Graph.from_edges(edges)

    # 0.2 * 45 = approx 9 edges expected
    n_edges = 0
    nnz = 0
    n_trials = 100
    for _ in range(n_trials):
        shortest_path_graph = preprocess.graph.shortest_paths(
            graph, retain_fraction=0.2
        )
        n_edges += shortest_path_graph.n_edges
        nnz += shortest_path_graph.A.nnz
    mean_n_edges = n_edges / n_trials
    mean_nnz = nnz / n_trials

    assert shortest_path_graph.n_items == n
    testing.assert_allclose(mean_n_edges, 9.0, atol=1.5)
    testing.assert_allclose(mean_nnz, 18.0, atol=1.5)

    assert (
        not (shortest_path_graph.A != shortest_path_graph.A.T).todense().all()
    )


def test_graph_from_numpy():
    X = np.ones(16).reshape(4, 4)
    np.fill_diagonal(X, 0.0)
    graph = preprocess.graph.Graph(X)
    assert graph.n_items == 4
    assert graph.n_edges == 6
    testing.assert_all_equal(torch.ones(graph.n_edges), graph.distances)

    expected_edges = torch.tensor(
        [[0, 1], [0, 2], [0, 3], [1, 2], [1, 3], [2, 3]]
    )
    testing.assert_all_equal(expected_edges, graph.edges)


def test_graph_from_torch():
    X = torch.ones(16).reshape(4, 4)
    X.fill_diagonal_(0.0)
    graph = preprocess.graph.Graph(X)
    assert graph.n_items == 4
    assert graph.n_edges == 6
    testing.assert_all_equal(torch.ones(graph.n_edges), graph.distances)

    expected_edges = torch.tensor(
        [[0, 1], [0, 2], [0, 3], [1, 2], [1, 3], [2, 3]]
    )
    testing.assert_all_equal(expected_edges, graph.edges)


def test_graph_from_csc_matrix():
    X = np.ones(16).reshape(4, 4)
    np.fill_diagonal(X, 0.0)
    X = sp.csc_matrix(X)
    with testing.disable_logging():
        # suppress a warning re: non-CSR matrix.
        graph = preprocess.graph.Graph(X)
    assert graph.n_items == 4
    assert graph.n_edges == 6
    testing.assert_all_equal(torch.ones(graph.n_edges), graph.distances)

    expected_edges = torch.tensor(
        [[0, 1], [0, 2], [0, 3], [1, 2], [1, 3], [2, 3]]
    )
    testing.assert_all_equal(expected_edges, graph.edges)



"""Datasets

Module for downloading datasets.

The functions in this module return a Dataset instance. Each function
takes at least two arguments:

    root:
        path to directory in which data should be downloaded (default './')
    download:
        bool indicating whether to download data if not present
        (default True)
"""

import pickle
import os

import numpy as np
from pymde.preprocess.graph import Graph
from pymde.problem import LOGGER
import scipy.sparse as sp
import torch
import torchvision


def _is_cached(root, files):
    return all(os.path.exists(os.path.join(root, fname)) for fname in files)


def _install_headers():
    # install simple header to bypass mod_security
    import urllib

    opener = urllib.request.build_opener()
    opener.addheaders = [("User-agent", "Mozilla/5.0")]
    urllib.request.install_opener(opener)


class Dataset(object):
    """Represents a dataset.

    Each instance has two main attrs:

        data:
            either a data matrix, or an instance of pymde.Graph; the
            data to be embedded.

        attributes:
            a dictionary whose values give attributes associated
            with the items in the data, such as the digit labels in MNIST.

    Other data that the dataset might carry is in the (dict) attribute
    `other_data`. Metadata about where the data came from is stored in
    `metadata`.
    """

    def __init__(self, data, attributes, other_data=None, metadata=None):
        self.data = data
        self.attributes = attributes
        self.other_data = other_data if other_data is not None else {}
        self.metadata = metadata if metadata is not None else {}

    @property
    def attribute(self):
        if len(self.attributes) == 1:
            return list(self.attributes.values())[0]


def FashionMNIST(root="./", download=True) -> Dataset:
    """Fashion-MNIST dataset (Xiao, et al.).

    The Fashion-MNIST dataset contains 70,000, 28x28 images of
    Zalando fashion articles.

    - ``data``: ``torch.Tensor`` with 70,000 rows, each of
      length 784 (representing the pixels in the image).
    - ``attributes`` dict: the key ``class`` holds an array
      in which each entry gives the fashion article in the corresponding row of
      ``data``.
    """
    root = os.path.expanduser(root)

    extract_root = os.path.join(root, "FashionMNIST")
    raw = os.path.join(extract_root, "raw")
    processed = os.path.join(extract_root, "processed")
    raw_files = [
        "t10k-images-idx3-ubyte.gz",
        "t10k-labels-idx1-ubyte.gz",
        "train-images-idx3-ubyte.gz",
        "train-labels-idx1-ubyte.gz",
    ]
    processed_files = ["test.pt", "training.pt"]

    def load_dataset():
        f_mnist_train = torchvision.datasets.FashionMNIST(
            root=root,
            train=True,
            download=True,
        )
        f_mnist_test = torchvision.datasets.FashionMNIST(
            root=root,
            train=False,
            download=True,
        )

        images = torch.cat([f_mnist_train.data, f_mnist_test.data])
        label = torch.cat([f_mnist_train.targets, f_mnist_test.targets])
        attributes = {"class": label}
        return Dataset(
            data=images.reshape(-1, 784),
            attributes=attributes,
            metadata={
                "authors": "Xiao, et al.",
                "url": "https://github.com/zalandoresearch/fashion-mnist",
            },
        )

    if _is_cached(raw, raw_files) and _is_cached(processed, processed_files):
        LOGGER.info("Loading cached dataset.")
        return load_dataset()

    if not download:
        raise RuntimeError("`download` is False, but data is not cached.")

    _install_headers()

    return load_dataset()


def MNIST(root="./", download=True) -> Dataset:
    """MNIST dataset (LeCun, et al.).

    The MNIST dataset contains 70,000, 28x28 images of handwritten digits.

    - ``data``: ``torch.Tensor`` with 70,000 rows, each of
      length 784 (representing the pixels in the image).
    - ``attributes`` dict: the key ``digits`` holds an array
      in which each entry gives the digit depicted in the corresponding row of
      ``data``.
    """
    url = "https://akshayagrawal.com/mnist/MNIST.tar.gz"

    root = os.path.expanduser(root)

    extract_root = os.path.join(root, "MNIST")
    raw = os.path.join(extract_root, "raw")
    processed = os.path.join(extract_root, "processed")
    raw_files = [
        "t10k-images-idx3-ubyte.gz",
        "t10k-labels-idx1-ubyte.gz",
        "train-images-idx3-ubyte.gz",
        "train-labels-idx1-ubyte.gz",
    ]
    processed_files = ["test.pt", "training.pt"]

    def load_dataset():
        mnist_train = torchvision.datasets.MNIST(
            root=root,
            train=True,
            download=False,
        )
        mnist_test = torchvision.datasets.MNIST(
            root=root,
            train=False,
            download=False,
        )

        images = torch.cat([mnist_train.data, mnist_test.data]).float()
        digits = torch.cat([mnist_train.targets, mnist_test.targets])
        attributes = {"digits": digits}
        return Dataset(
            data=images.reshape(-1, 784),
            attributes=attributes,
            metadata={
                "authors": "LeCunn, et al.",
                "url": "http://yann.lecun.com/exdb/mnist/",
            },
        )

    if _is_cached(raw, raw_files) and _is_cached(processed, processed_files):
        LOGGER.info("Loading cached dataset.")
        return load_dataset()

    if not download:
        raise RuntimeError("`download` is False, but data is not cached.")

    _install_headers()
    filename = url.rpartition("/")[2]
    LOGGER.info("Downloading MNIST dataset ...")
    torchvision.datasets.utils.download_and_extract_archive(
        url, download_root=root, filename=filename
    )
    os.remove(os.path.join(root, filename))
    LOGGER.info("Download complete.")

    return load_dataset()


def google_scholar(root="./", download=True, full=False) -> Dataset:
    """Google Scholar dataset (Agrawal, et al.).

    The Google Scholar dataset contains an academic coauthorship graph: the
    nodes are authors, and two authors are connected by an edge if either
    author listed the other as a coauthor on Google Scholar. (Note that if
    two authors collaborated on a paper, but neither has listed the other
    as a coauthor on their Scholar profiles, then they will not be connected
    by an edge).

    If ``full`` is False, obtains a small version of the dataset, on roughly
    40,000 authors, each with h-index at least 50. If ``full`` is True,
    obtains the whole dataset, on roughly 600,000 authors. The full dataset
    is roughly 1GB in size.

    - ``data``: a ``pymde.Graph``, with nodes representing authors
    - ``attributes``: the ``coauthors`` key has an array holding the number
      of coauthors of each other, normalized to be a percentile.
    - ``other_data``: holds a dataframe describing the dataset, keyed by
      ``dataframe``.
    """

    try:
        import pandas as pd
    except ImportError:
        raise ImportError("Please install pandas.")

    root = os.path.expanduser(root)

    if not full:
        url = "https://akshayagrawal.com/scholar/google_scholar.tar.gz"
    else:
        url = "https://akshayagrawal.com/scholar/google_scholar_full.tar.gz"

    extract_root = os.path.join(root, "scholar/")
    data_file = os.path.join(extract_root, "scholar_graph.npz")
    dataframe_file = os.path.join(extract_root, "scholar_df.pkl")
    if not full:
        disciplines_dataframe_file = os.path.join(
            extract_root, "scholar_disciplines_df.pkl"
        )

    def load_dataset():
        data = Graph(sp.load_npz(os.path.join(root, data_file)))
        df = pd.read_pickle(os.path.join(root, dataframe_file))
        attributes = {"coauthors": df["coauthors_pct"]}
        if not full:
            disciplines_df = pd.read_pickle(
                os.path.join(root, disciplines_dataframe_file)
            )
            other_data = {"dataframe": df, "disciplines": disciplines_df}
        else:
            other_data = None
        metadata = {"authors": "Agrawal"}
        return Dataset(data, attributes, other_data, metadata)

    if not full:
        files = [
            data_file,
            dataframe_file,
            disciplines_dataframe_file,
        ]
    else:
        files = [
            data_file,
            dataframe_file,
        ]

    if _is_cached(root, files):
        LOGGER.info("Loading cached dataset.")
        return load_dataset()

    if not download:
        raise RuntimeError("`download` is False, but data is not cached.")

    _install_headers()

    filename = url.rpartition("/")[2]
    torchvision.datasets.utils.download_and_extract_archive(
        url, download_root=root, extract_root=extract_root, filename=filename
    )
    os.remove(os.path.join(root, filename))
    LOGGER.info("Dataset is now available.")
    return load_dataset()


def covid19_scrna_wilk(root="./", download=True) -> Dataset:
    """COVID-19 scRNA data (Wilk, et al.).

    The COVID-19 dataset includes a PCA embedding of single-cell
    mRNA transcriptomes of roughly 40,000 cells, taken from some patients
    with COVID-19 infections and from healthy controls.

    Instructions on how to obtain the full dataset are available in the
    Wilk et al. paper: https://www.nature.com/articles/s41591-020-0944-y,

    - ``data``: the PCA embedding
    - ``attributes``: two keys, ``cell_type`` and ``health_status``.
    """
    root = os.path.expanduser(root)

    url = "https://akshayagrawal.com/scrna/scrna_covid19_wilk.tar.gz"

    extract_root = os.path.join(root, "scrna_covid19_wilk/")
    data_file = os.path.join(extract_root, "scrna_data_matrix.npy")
    cell_type_attr_file = os.path.join(extract_root, "scrna_cell_type_attr.npy")
    health_status_attr_file = os.path.join(
        extract_root, "scrna_health_status_attr.npy"
    )

    def load_dataset():
        data = np.load(os.path.join(root, data_file))
        attributes = {
            "cell_type": np.load(
                os.path.join(root, cell_type_attr_file), allow_pickle=True
            ),
            "health_status": np.load(
                os.path.join(root, health_status_attr_file), allow_pickle=True
            ),
        }
        metadata = {
            "authors": "Wilk, et al.",
            "url": "https://www.nature.com/articles/s41591-020-0944-y",
        }
        return Dataset(data, attributes, metadata=metadata)

    files = [
        data_file,
        cell_type_attr_file,
        health_status_attr_file,
    ]

    if _is_cached(root, files):
        LOGGER.info("Loading cached dataset.")
        return load_dataset()

    if not download:
        raise RuntimeError("`download` is False, but data is not cached.")

    _install_headers()

    filename = url.rpartition("/")[2]
    torchvision.datasets.utils.download_and_extract_archive(
        url, download_root=root, extract_root=extract_root, filename=filename
    )
    os.remove(os.path.join(root, filename))
    LOGGER.info("Dataset is now available.")
    return load_dataset()


def population_genetics(root="./", download=True) -> Dataset:
    """Population genetics dataset (Nelson, et al)

    The population genetics dataset includes a PCA embedding (in R^20) of
    single nucleotide polymorphism data associated with 1,387
    individuals thought to be of European descent. (The data is from the
    Population Reference Sample project by Nelson, et al.)

    It also contains a "corrupted" version of the data, in which 154 additional
    points have been injected; the first 10 coordinates of these synthetic
    points are generated using a discrete uniform distribution on
    {0, 1, 2}, and the last 10 are generated using a discrete uniform
    distritubtion on {1/12, /18}.

    A study of Novembre et al (2008) showed that a PCA embedding in R^2
    roughly resembles the map of Europe, suggesting that the genes
    encode geographical information. But PCA does not produce interesting
    visualizations of the corrupted data. If distortion functions are chosen to
    be robust (eg, using the Log1p or Huber attractive penalties), we can
    create embeddings that preserve the geographical structure, while
    placing the synthetic points to the side, in their own cluser.

    - ``data``: the PCA embedding of the clean genetic data, in R^20
    - ``corrupted_data``: the corrupted data, in R^20
    - ``attributes``: two keys, ``clean_colors`` and ``corrupted_colors``.
    """
    try:
        import pandas as pd
    except ImportError:
        raise ImportError("Please install pandas.")

    root = os.path.expanduser(root)

    url = "https://akshayagrawal.com/popres/popres.tar.gz"

    extract_root = os.path.join(root, "population_genetics/")
    data_file = os.path.join(extract_root, "data.csv")
    corrupted_data_file = os.path.join(extract_root, "corrupted_data.csv")
    clean_attr_file = os.path.join(extract_root, "data_colors.csv")
    corrupted_attr_file = os.path.join(extract_root, "corrupted_colors.csv")

    def load_dataset():
        data = pd.read_csv(os.path.join(root, data_file)).to_numpy()
        corrupted_data = pd.read_csv(
            os.path.join(root, corrupted_data_file)
        ).to_numpy()
        attributes = {
            "clean_colors": pd.read_csv(
                os.path.join(root, clean_attr_file)
            ).to_numpy(),
            "corrupted_colors": pd.read_csv(
                os.path.join(root, corrupted_attr_file)
            ).to_numpy(),
        }
        metadata = {
            "authors": "Nelson, et al",
            "url": "https://pubmed.ncbi.nlm.nih.gov/18760391/",
        }
        dataset = Dataset(data, attributes, metadata=metadata)
        dataset.corrupted_data = corrupted_data
        return dataset

    files = [
        data_file,
        corrupted_data_file,
        clean_attr_file,
        corrupted_attr_file,
    ]

    if _is_cached(root, files):
        LOGGER.info("Loading cached dataset.")
        return load_dataset()

    if not download:
        raise RuntimeError("`download` is False, but data is not cached.")

    _install_headers()

    filename = url.rpartition("/")[2]
    torchvision.datasets.utils.download_and_extract_archive(
        url, download_root=root, extract_root=extract_root, filename=filename
    )
    os.remove(os.path.join(root, filename))
    LOGGER.info("Dataset is now available.")
    return load_dataset()


def counties(root="./", download=True) -> Dataset:
    """US counties (2013-2017 ACS 5-Year Estimates)

    This dataset contains 34 demographic features for each of the 3,220 US
    counties. The features were collected by the 2013-2017 ACS 5-Year
    Estimates longitudinal survey, run by the US Census Bureau.

    - ``data``: the PCA embedding of the clean genetic data, in R^20
    - ``county_dataframe``: the raw ACS data
    - ``voting_dataframe``: the raw 2016 voting data
    - ``attributes``: one key, ``democratic_fraction``, the fraction of
      of voters who voted Democratic in each county in the 2016 presidential
      election
    """
    try:
        import pandas as pd
    except ImportError:
        raise ImportError("Please install pandas.")

    root = os.path.expanduser(root)

    url = "https://akshayagrawal.com/voting/voting.tar.gz"

    extract_root = os.path.join(root, "counties/")
    data_file = os.path.join(extract_root, "county_data.npy")
    voting_attr_file = os.path.join(extract_root, "democratic_fraction.npy")

    county_df_file = os.path.join(extract_root, "acs2017_county_data.csv")
    voting_df_file = os.path.join(
        extract_root, "2016_US_County_Level_Presidential_Results.csv"
    )

    def load_dataset():
        data = np.load(os.path.join(root, data_file), allow_pickle=True)
        county_dataframe = pd.read_csv(os.path.join(root, county_df_file))
        voting_dataframe = pd.read_csv(os.path.join(root, voting_df_file))
        attributes = {
            "democratic_fraction": np.load(os.path.join(root, voting_attr_file))
        }
        metadata = {
            "authors": "ACS 2013-2017 survey",
            "url": "https://www.census.gov/newsroom/press-kits/2018/acs-5year.html",  # noqa: E501
        }
        dataset = Dataset(data, attributes, metadata=metadata)
        dataset.county_dataframe = county_dataframe
        dataset.voting_dataframe = voting_dataframe
        return dataset

    files = [
        data_file,
        voting_attr_file,
        county_df_file,
        voting_attr_file,
    ]

    if _is_cached(root, files):
        LOGGER.info("Loading cached dataset.")
        return load_dataset()

    if not download:
        raise RuntimeError("`download` is False, but data is not cached.")

    _install_headers()

    filename = url.rpartition("/")[2]
    torchvision.datasets.utils.download_and_extract_archive(
        url, download_root=root, extract_root=extract_root, filename=filename
    )
    os.remove(os.path.join(root, filename))
    LOGGER.info("Dataset is now available.")
    return load_dataset()


def google_scholar_interests(root="./", download=True) -> Dataset:
    """Google Scholar academic interests

    This dataset contains a cooccurrence matrix of the 5000 most popular
    academic interests listed by authors on Google Scholar

    - ``data``: a cooccurrence matrix, counting the number of times two
      two interests appeared among a single author's listed interests,
      with ``data[i, j]`` giving the cooccurrence count of ``interests[i]`` and
      ``interests[j]``
    - ``attributes``: one key, ``interests``, where ``interests[i]`` is the
      interest corresponding to the ith row/column of ``data``.
    """
    root = os.path.expanduser(root)

    url = "https://akshayagrawal.com/scholar/interests.tar.gz"

    extract_root = os.path.join(root, "interests/")
    data_file = os.path.join(extract_root, "topic_cooccurrence_matrix.npz")
    interests_attr_file = os.path.join(extract_root, "interests.pkl")

    def load_dataset():
        data = Graph(sp.load_npz(data_file))
        with open(interests_attr_file, "rb") as f:
            interests = pickle.load(f)
        attributes = {"interests": interests}
        metadata = {
            "authors": "Agrawal",
        }
        dataset = Dataset(data, attributes, metadata=metadata)
        return dataset

    files = [
        data_file,
        interests_attr_file,
    ]

    if _is_cached(root, files):
        LOGGER.info("Loading cached dataset.")
        return load_dataset()

    if not download:
        raise RuntimeError("`download` is False, but data is not cached.")

    _install_headers()

    filename = url.rpartition("/")[2]
    torchvision.datasets.utils.download_and_extract_archive(
        url, download_root=root, extract_root=extract_root, filename=filename
    )
    os.remove(os.path.join(root, filename))
    LOGGER.info("Dataset is now available.")
    return load_dataset()



__version__ = "0.1.16"

from pymde.problem import MDE

from pymde.constraints import Centered, Anchored, Standardized

from pymde import datasets

from pymde.preprocess.graph import Graph

from pymde.util import all_edges
from pymde.util import align
from pymde.util import center
from pymde.util import rotate
from pymde.util import seed

from pymde.experiment_utils import latexify, plot

from pymde.functions import losses, penalties

from pymde.quadratic import pca
from pymde.recipes import (
    laplacian_embedding,
    preserve_distances,
    preserve_neighbors,
)



import numpy as np
import pytest
import torch

from pymde import problem
from pymde import quadratic
from pymde import util
from pymde.constraints import Standardized
from pymde.functions import penalties
import pymde.testing as testing


@testing.cpu_and_cuda
def test_initialization(device):
    torch.random.manual_seed(0)
    constraint = Standardized()
    X = constraint.initialization(5, 3, device=device)
    testing.assert_allclose(1.0 / 5 * X.T @ X, np.eye(3))


@testing.cpu_and_cuda
def test_differences(device):
    torch.random.manual_seed(0)
    edges = np.array([(0, 1), (0, 2), (1, 2)])
    X = torch.randn((3, 3), dtype=torch.float32, device=device)
    mde = problem.MDE(
        3,
        3,
        edges,
        penalties.Quadratic(torch.ones(3)),
        constraint=Standardized(),
        device=device,
    )
    diff = mde.differences(X)
    testing.assert_allclose(X[edges[:, 0]] - X[edges[:, 1]], diff)


@testing.cpu_and_cuda
def test_norms(device):
    torch.random.manual_seed(0)
    edges = np.array([(0, 1), (0, 2), (1, 2)])
    X = torch.randn((3, 3), dtype=torch.float32, device=device)
    mde = problem.MDE(
        3,
        3,
        edges,
        penalties.Quadratic(torch.ones(3)),
        constraint=Standardized(),
        device=device,
    )
    diff = X[edges[:, 0]] - X[edges[:, 1]]
    norms = diff.norm(dim=1)
    testing.assert_allclose(norms, mde.distances(X))


@testing.cpu_and_cuda
def test_norm_grad_zero(device):
    torch.random.manual_seed(0)
    edges = np.array([(0, 1)])
    mde = problem.MDE(
        3,
        3,
        edges,
        penalties.Quadratic(torch.ones(3)),
        constraint=Standardized(),
        device=device,
    )
    X = torch.ones((3, 3), requires_grad=True, device=device)
    norms = mde.distances(X)
    norms.backward()
    testing.assert_allclose(X.grad, 0.0)


@testing.cpu_and_cuda
def test_average_distortion(device):
    torch.random.manual_seed(0)
    edges = np.array([(0, 1), (0, 2), (1, 2)])
    mde = problem.MDE(
        3,
        2,
        edges,
        penalties.Quadratic(torch.tensor([1.0, 2.0, 3.0])),
        constraint=Standardized(),
        device=device,
    )
    X = torch.tensor(
        [[0.0, 0.0], [1.0, 1.0], [3.0, 3.0]],
        dtype=torch.float32,
        device=device,
    )
    average_distortion = mde.average_distortion(X)
    # (1*2 + 2*18 + 3*8)/3 = (2 + 36 + 24)/3 = 62/3
    testing.assert_allclose(average_distortion.detach().cpu().numpy(), 62.0 / 3)


@testing.cpu_and_cuda
def test_average_distortion_grad(device):
    torch.random.manual_seed(0)
    edges = np.array([(0, 1), (0, 2), (1, 2)])
    f = penalties.Quadratic(torch.tensor([1.0, 2.0, 3.0], device=device))
    mde = problem.MDE(3, 2, edges, f, Standardized(), device=device)
    X = torch.randn(
        (3, 2),
        requires_grad=True,
        dtype=torch.float32,
        device=device,
    )
    average_distortion = mde.average_distortion(X)
    average_distortion.backward()
    A = torch.tensor(
        [[1, 1, 0], [-1, 0, 1], [0, -1, -1]],
        device=device,
    ).float()
    auto_grad = X.grad
    X.grad = None
    util._distortion(X, f, A, mde._lhs, mde._rhs).backward()
    manual_grad = X.grad
    testing.assert_allclose(auto_grad, manual_grad)


@testing.cpu_and_cuda
def test_fit_spectral(device):
    # TODO deflake this test
    pytest.skip("This test is flaky on macOS.")
    util.seed(0)
    n = 200
    m = 3
    max_iter = 1000

    edges = util.all_edges(n)
    weights = torch.ones(edges.shape[0])
    f = penalties.Quadratic(weights)

    mde = problem.MDE(
        n,
        m,
        edges=edges,
        distortion_function=f,
        constraint=Standardized(),
        device=device,
    )
    X = mde.embed(max_iter=max_iter, eps=1e-10, memory_size=10)

    assert id(X) == id(mde.X)
    X_spectral = quadratic.spectral(
        n, m, edges=edges, weights=weights, device=device
    )

    testing.assert_allclose(
        mde.average_distortion(X).detach().cpu().numpy(),
        mde.average_distortion(X_spectral).detach().cpu().numpy(),
        atol=1e-4,
    )


@testing.cpu_and_cuda
def test_self_edges_raises_error(device):
    torch.random.manual_seed(0)
    edges = np.array([(0, 1), (0, 0), (0, 2), (1, 2), (1, 1)])
    with pytest.raises(
        ValueError, match=r"The edge list must not contain self edges.*"
    ):
        problem.MDE(
            3,
            3,
            edges,
            penalties.Quadratic(torch.ones(edges.shape[0])),
            constraint=Standardized(),
            device=device,
        )



import numpy as np
import scipy.sparse as sp
import torch

from pymde import problem
from pymde.preprocess.graph import Graph
from pymde.preprocess.preprocess import sample_edges
from pymde import util


def distances(data, retain_fraction=1.0, verbose=False):
    """Compute distances, given data matrix.

    This function computes distances between some pairs of items, given
    a data matrix. Each row in the data matrix is treated as an item.

    Arguments
    ---------
    data: torch.Tensor, np.ndarray, or scipy.sparse matrix
        The data matrix, shape ``(n_items, n_features)``.
    retain_fraction: float, optional
        A float between 0 and 1, specifying the fraction of all ``(n_items
        choose 2)`` to compute. For example, if ``retain_fraction`` is 0.1,
        only 10 percent of the edges will be stored.
    verbose:
        If ``True``, print verbose output.

    Returns
    -------
    pymde.Graph
        A graph object holding the distances and corresponding edges.
        Access the distances with ``graph.distances``, and the edges
        with ``graph.edges``.
    """
    if not sp.issparse(data) and not isinstance(
        data, (np.ndarray, torch.Tensor)
    ):
        raise ValueError(
            "`data` must be a scipy.sparse matrix, NumPy array, "
            "or torch tensor"
        )

    n_items = int(data.shape[0])
    all_edges = n_items * (n_items - 1) / 2
    max_distances = int(retain_fraction * all_edges)

    if max_distances is None:
        max_distances = np.inf
    elif max_distances <= 0:
        raise ValueError("max_distances must be positive")

    if n_items * (n_items - 1) / 2 < max_distances:
        edges = util.all_edges(n_items)
    else:
        if verbose:
            problem.LOGGER.info(f"Sampling {int(max_distances)} edges")
        edges = sample_edges(n_items, int(max_distances))

    if sp.issparse(data):
        if not isinstance(data, sp.csr_matrix):
            data = data.tocsr()

        edges = edges.cpu().numpy()
        if verbose:
            problem.LOGGER.info(f"Computing {int(edges.shape[0])} distances")
        delta = torch.tensor(
            sp.linalg.norm(data[edges[:, 0]] - data[edges[:, 1]], axis=1),
            dtype=torch.float,
        )
        edges = torch.tensor(edges)
    elif isinstance(data, (np.ndarray, torch.Tensor)):
        if isinstance(data, np.ndarray):
            data = torch.tensor(data, dtype=torch.float, device="cpu")

        edges = edges.to(data.device)
        if verbose:
            problem.LOGGER.info(f"Computing {int(edges.shape[0])} distances")
        # TODO(akshayka): Batch this computation when the number of edges
        # and/or the number of features is large.
        delta = (
            (data[edges[:, 0]] - data[edges[:, 1]])
            .pow(2)
            .sum(dim=1)
            .float()
            .sqrt()
        )

    return Graph.from_edges(edges, delta, n_items=n_items)


def k_nearest_neighbors(data, k, max_distance=None, verbose=False):
    """Compute k-nearest neighbors for each row in data matrix.

    Computes the k-nearest neighbor graph of data matrix, under
    the Euclidean distance. Each row in the data matrix is treated as an item.

    Arguments
    ---------
    data: {torch.Tensor, np.ndarray, scipy.sparse matrix}(
            shape=(n_items, n_features))
        The data matrix
    k: int
        The number of nearest neighbors per item
    max_distance: float (optional)
        If not None, neighborhoods are restricted to have a radius
        no greater than `max_distance`.
    verbose: bool
        If True, print verbose output.

    Returns
    -------
    pymde.Graph
        a neighborhood graph
    """
    # lazy import, because importing pynndescent takes some time
    import pynndescent

    if isinstance(data, torch.Tensor):
        device = data.device
        data = data.cpu().numpy()
    else:
        device = "cpu"

    n = data.shape[0]
    if n < 10000:
        import sklearn.neighbors

        if verbose:
            problem.LOGGER.info("Exact nearest neighbors by brute force ")
        nn = sklearn.neighbors.NearestNeighbors(
            n_neighbors=k + 1, algorithm="brute"
        )
        nn.fit(data)
        distances, neighbors = nn.kneighbors(data)
    else:
        # TODO default params (n_trees, max_candidates)
        index = pynndescent.NNDescent(
            data,
            n_neighbors=k + 1,
            verbose=verbose,
            max_candidates=60,
        )
        neighbors, distances = index.neighbor_graph
    neighbors = neighbors[:, 1:]
    distances = distances[:, 1:]

    n = data.shape[0]
    items = np.arange(n)
    items = np.repeat(items, k)
    edges = np.stack([items, neighbors.flatten()], axis=1)

    flip_idx = edges[:, 0] > edges[:, 1]
    edges[flip_idx] = np.stack(
        [edges[flip_idx][:, 1], edges[flip_idx][:, 0]], axis=1
    )

    duplicated_edges_mask = edges[:, 0] == edges[:, 1]
    if duplicated_edges_mask.any():
        problem.LOGGER.warning(
            "Your dataset appears to contain duplicated items (rows); "
            "when embedding, you should typically have unique items."
        )
        problem.LOGGER.warning(
            "The following items have duplicates "
            f"{edges[duplicated_edges_mask][:, 0]}"
        )
        edges = edges[~duplicated_edges_mask]

    weights = torch.ones(edges.shape[0], device=device, dtype=torch.float)
    if max_distance is not None:
        weights[
            torch.tensor(distances.ravel(), device=device, dtype=torch.float)
            > max_distance
        ] = 0.0

    # weights for duplicated edges will be summed.
    edges = torch.tensor(edges, device=device)
    return Graph.from_edges(edges, weights)


# TODO(akshayka) figure out this api ...
def _neighborhood_graph(
    data, n_neighbors=None, threshold=None, max_distances=None
):
    if n_neighbors is None and threshold is None:
        n_neighbors = 15
    elif n_neighbors is not None and threshold is not None:
        raise ValueError(
            "only one of n_neighbors and threshold can be non-none"
        )

    if n_neighbors is not None:
        return k_nearest_neighbors(data, n_neighbors)
    else:
        # TODO: move from recipes to here
        raise NotImplementedError


def _distances(index):
    pass


def _distance_matrix(data, max_distances=None):
    """Compute a distance matrix from a data matrix"""
    # TODO(akshayka): move from recipes to here
    # return vector of distances/edges, or a graph, or distance matrix?
    raise NotImplementedError



import numpy as np
import pytest
import torch

from pymde import util
import pymde.testing as testing


def test_to_tensor():
    args = [0.0, np.array(1.0), torch.tensor(2.0)]
    tensors = util.to_tensor(args)
    for t, arg in zip(tensors, args):
        assert isinstance(t, torch.Tensor)
        assert t.item() == float(arg)
        assert str(t.device) == "cpu"
    assert id(tensors[-1]) == id(args[-1])


@testing.cpu_and_cuda
def test_proj_standardized(device):
    X = torch.eye(2, dtype=torch.float32, device=device)
    proj = util.proj_standardized(X)
    testing.assert_allclose(1 / 2.0 * proj.T @ proj, np.eye(2))

    n = 10
    m = 3
    X = torch.randn((n, m), dtype=torch.float32, device=device)
    proj = util.proj_standardized(X)
    testing.assert_allclose(1.0 / n * proj.T @ proj, np.eye(m))

    n = 100
    m = 3
    X = torch.randn((n, m), dtype=torch.float32, device=device)
    proj = util.proj_standardized(X)
    testing.assert_allclose(1.0 / n * proj.T @ proj, np.eye(m))

    n = 100
    m = 3
    X = torch.randn((n, m), dtype=torch.float32, device=device)
    X -= X.mean(axis=0)
    proj = util.proj_standardized(X)
    testing.assert_allclose(1.0 / n * proj.T @ proj, np.eye(m))
    testing.assert_allclose(proj.mean(axis=0), np.zeros(m))

    n = 100
    m = 3
    X = torch.randn((n, m), dtype=torch.float32, device=device)
    proj = util.proj_standardized(X, demean=True)
    testing.assert_allclose(1.0 / n * proj.T @ proj, np.eye(m))
    testing.assert_allclose(proj.mean(axis=0), np.zeros(m))

    n = 1000
    m = 2
    X = torch.randn((n, m), dtype=torch.float32, device=device)
    proj = util.proj_standardized(X, demean=True)
    testing.assert_allclose(1.0 / n * proj.T @ proj, np.eye(m))
    testing.assert_allclose(proj.mean(axis=0), np.zeros(m))

    n = 1000
    m = 3
    X = torch.randn((n, m), dtype=torch.float32, device=device)
    proj = util.proj_standardized(X, demean=True)
    testing.assert_allclose(1.0 / n * proj.T @ proj, np.eye(m))
    testing.assert_allclose(proj.mean(axis=0), np.zeros(m))

    n = 1000
    m = 250
    X = torch.randn((n, m), dtype=torch.float32, device=device)
    proj = util.proj_standardized(X, demean=True)
    testing.assert_allclose(1.0 / n * proj.T @ proj, np.eye(m))
    testing.assert_allclose(proj.mean(axis=0), np.zeros(m))


def test_rng():
    util.seed(0)
    tensor = torch.randn((10, 5))
    from_random_state = np.random.randn(10, 5)
    from_rng = util.np_rng().standard_normal((10, 5))

    util.seed(0)
    testing.assert_all_equal(tensor, torch.randn((10, 5)))
    testing.assert_all_equal(from_random_state, np.random.randn(10, 5))
    testing.assert_all_equal(from_rng, util.np_rng().standard_normal((10, 5)))


def test_align():
    pytest.skip("Unimplemented.")


def test_adjacency_matrix():
    pytest.skip("Unimplemented.")



cimport numpy as cnp


ctypedef int ITYPE_t

# NULL_IDX is the index used in predecessor matrices to store a non-path
DEF NULL_IDX = -9999


def _breadth_first_directed(
                           unsigned int head_node,
                           cnp.ndarray[ITYPE_t, ndim=1, mode='c'] indices,
                           cnp.ndarray[ITYPE_t, ndim=1, mode='c'] indptr,
                           cnp.ndarray[ITYPE_t, ndim=1, mode='c'] node_list,
                           cnp.ndarray[ITYPE_t, ndim=1, mode='c'] lengths,
                           cnp.ndarray[ITYPE_t, ndim=1, mode='c'] predecessors):
    # Inputs:
    #  head_node: (input) index of the node from which traversal starts
    #  indices: (input) CSR indices of graph
    #  indptr:  (input) CSR indptr of graph
    #  node_list: (output) breadth-first list of nodes
    #  predecessors: (output) list of predecessors of nodes in breadth-first
    #                tree.  Should be initialized to NULL_IDX
    # Returns:
    #  n_nodes: the number of nodes in the breadth-first tree
    cdef unsigned int i, pnode, cnode
    cdef unsigned int i_nl, i_nl_end
    cdef unsigned int N = node_list.shape[0]
    cdef unsigned int curr_length

    node_list[0] = head_node
    lengths[head_node] = 0
    i_nl = 0
    i_nl_end = 1

    while i_nl < i_nl_end:
        pnode = node_list[i_nl]
        curr_length = lengths[pnode]

        for i in range(indptr[pnode], indptr[pnode + 1]):
            cnode = indices[i]
            if (cnode == head_node):
                continue
            elif (predecessors[cnode] == NULL_IDX):
                node_list[i_nl_end] = cnode
                predecessors[cnode] = pnode
                lengths[cnode] = curr_length + 1
                i_nl_end += 1

        i_nl += 1

    return i_nl





"""Recipes for constructing MDE problems.
"""
import numpy as np
import scipy.sparse
import torch


from pymde import problem
from pymde import quadratic
from pymde import constraints
from pymde.functions import penalties, losses
from pymde import preprocess


def _remove_anchor_anchor_edges(edges, data, anchors, opt=None, blk=1e8):
    """
    Exclude edges in which both items are anchors, since these
    items are already pinned in place by the anchor constraint

    TODO(akshayka): handle edge case in which every item is anchored,
    in which case the edge list will be empty after this exclusion

    Arguments
    ---------
    edges:       edge list N x 2
    data:        data to filter  N x ?
    anchors:     remove anchor-to-anchor edges from edges[] and data[]
    tuning parameters [optional, for fine-tuning the heuristics]
      opt:       [None=automatic opt & blk] which implementation?
      blk:       loop-blocking heuristic mem limit

    `edges` and `data` can be torch or numpy arrays

    Returns
    -------
    filtered versions of edges & data
    """
    na = anchors.shape[0]
    if na <= 0:
        return edges, data
    blk = int(blk)
    ne = edges.shape[0]
    ea = (
        ne * na
    )  # torch uses "ea" bytes of tmp mem for opt==0 (maybe too large)

    if opt is None:  # set both opt and blk according to simple heuristic
        opt = 0
        blk = int(1e8)
        if ea > blk:  # opt=0 is a possibility
            opt = 1  # fall-back to 3
            if ne >= 150000:
                blk = 262144

    if opt == 0 and ea < blk:
        neither_anchors_mask = ~(
            (edges[:, 0][..., None] == anchors).any(-1)
            * (edges[:, 1][..., None] == anchors).any(-1)
        )

    elif opt == 1 and na < blk:  # no need to block anchor-loop
        # astep=na
        estep = int(blk / na)
        assert estep > 0

        msk0 = torch.zeros(ne, dtype=torch.bool, device=edges.device)
        msk1 = torch.zeros(ne, dtype=torch.bool, device=edges.device)
        e = 0
        while e < ne:
            ehi = e + estep if e + estep < ne else ne
            msk0[e:ehi] |= (edges[e:ehi, 0][..., None] == anchors).any(-1)
            msk1[e:ehi] |= (edges[e:ehi, 1][..., None] == anchors).any(-1)
            e += estep
        neither_anchors_mask = ~(msk0 & msk1)  # neither

    else:  # o.w block both edge and anchor loops
        astep = min(na, blk)
        estep = int(blk / astep)
        assert astep > 0
        assert estep > 0

        msk0 = torch.zeros(ne, dtype=torch.bool, device=edges.device)
        msk1 = torch.zeros(ne, dtype=torch.bool, device=edges.device)
        e = 0
        while e < ne:
            ehi = e + estep if e + estep < ne else ne
            e0 = edges[e:ehi, 0][..., None]
            e1 = edges[e:ehi, 1][..., None]
            a = 0
            while a < na:
                ahi = a + astep if a + astep < na else na
                msk0[e:ehi] |= (e0 == anchors[a:ahi]).any(-1)
                msk1[e:ehi] |= (e1 == anchors[a:ahi]).any(-1)
                a += astep
            e += estep
        neither_anchors_mask = ~(msk0 & msk1)  # neither

    edges = edges[neither_anchors_mask]
    data = data[neither_anchors_mask]
    return edges, data


def preserve_distances(
    data,
    embedding_dim=2,
    loss=losses.Absolute,
    constraint=None,
    max_distances=5e7,
    device="cpu",
    verbose=False,
) -> problem.MDE:
    """Construct an MDE problem based on original distances.

    This function constructs an MDE problem for preserving pairwise
    distances between items. This can be useful for preserving the global
    structure of the data.

    The data can be specified with either a data matrix (a NumPy array, torch
    Tensor, or sparse matrix), or a ``pymde.Graph`` instance encoding the
    distances:

        A NumPy array, torch tensor, or sparse matrix is interpreted as a
        collection of feature vectors: each row gives the feature vector for an
        item. The original distances are the Euclidean distances between the
        feature vectors.

        A ``pymde.Graph`` instance is interpreted as encoding all (n_items
        choose 2) distances: the distance between i and j is taken to be the
        length of the shortest path connecting i and j.

    When the number of items n_items is large, the total number of pairs will
    be very large. When this happens, instead of computing all pairs of
    distances, this function will sample a subset uniformly at random. The
    maximum number of distances to compute is specified by the parameter
    ``max_distances``. Depending on how many items you have (and how much
    memory your machine has), you may need to adjust this parameter.

    To obtain an embedding, call the ``embed`` method on the returned object.
    To plot it, use ``pymde.plot``.

    For example:

    .. code:: python3

        embedding = pymde.preserve_distances(data).embed()
        pymde.plot(embedding)

    Arguments
    ---------
    data: {np.ndarray, torch.Tensor, scipy.sparse matrix} or pymde.Graph
        The original data, a data matrix of shape ``(n_items, n_features)`` or
        a graph.
    embedding_dim: int
        The embedding dimension.
    loss: pymde.Function class (or factory)
        Callable that constructs a distortion function, given
        original distances. Typically one of the classes defined in
        ``pymde.losses``, such as ``pymde.losses.Absolute``, or
        ``pymde.losses.WeightedQuadratic``.
    constraint: pymde.constraints.Constraint (optional)
        Embedding constraint, such as ``pymde.Standardized()`` or
        ``pymde.Anchored(anchors, values)`` (or ``None``). Defaults to no
        constraint. Note: when the constraint is ``pymde.Standardized()``,
        the original distances will be scaled by a constant (because the
        standardization constraint puts a limit on how large any one
        distance can be).
    max_distances: int
        Maximum number of distances to compute.
    device: str (optional)
        Device for the embedding (eg, 'cpu', 'cuda').
    verbose: bool
        If ``True``, print verbose output.

    Returns
    -------
    pymde.MDE
        A ``pymde.MDE`` instance, based on preserving the original distances.
    """
    if not isinstance(
        data, (np.ndarray, torch.Tensor, preprocess.graph.Graph)
    ) and not scipy.sparse.issparse(data):
        raise ValueError(
            "`data` must be a np.ndarray/torch.Tensor/scipy.sparse matrix"
            ", or a pymde.Graph."
        )

    if isinstance(data, preprocess.graph.Graph):
        n_items = data.n_items
    else:
        n_items = data.shape[0]
    n_all_edges = (n_items) * (n_items - 1) / 2
    retain_fraction = max_distances / n_all_edges

    graph = preprocess.generic.distances(
        data, retain_fraction=retain_fraction, verbose=verbose
    )
    edges = graph.edges.to(device)
    deviations = graph.distances.to(device)

    if constraint is None:
        constraint = constraints.Centered()
    elif isinstance(constraint, constraints._Standardized):
        deviations = preprocess.scale(
            deviations, constraint.natural_length(n_items, embedding_dim)
        )
    elif isinstance(constraint, constraints.Anchored):
        edges, deviations = _remove_anchor_anchor_edges(
            edges, deviations, constraint.anchors
        )

    return problem.MDE(
        n_items=n_items,
        embedding_dim=embedding_dim,
        edges=edges,
        distortion_function=loss(deviations),
        constraint=constraint,
        device=device,
    )


def preserve_neighbors(
    data,
    embedding_dim=2,
    attractive_penalty=penalties.Log1p,
    repulsive_penalty=penalties.Log,
    constraint=None,
    n_neighbors=None,
    repulsive_fraction=None,
    max_distance=None,
    init="quadratic",
    device="cpu",
    verbose=False,
) -> problem.MDE:
    """Construct an MDE problem designed to preserve local structure.

    This function constructs an MDE problem for preserving the
    local structure of original data. This MDE problem is well-suited for
    visualization (using ``embedding_dim`` 2 or 3), but can also be used to
    generate features for machine learning tasks (with ``embedding_dim`` = 10,
    50, or 100, for example). It yields embeddings in which similar items
    are near each other, and dissimilar items are not near each other.

    The original data can either be a data matrix, or a graph.
    Data matrices should be torch Tensors, NumPy arrays, or scipy sparse
    matrices; graphs should be instances of ``pymde.Graph``.

    The MDE problem uses distortion functions derived from weights (i.e.,
    penalties).

    To obtain an embedding, call the ``embed`` method on the returned ``MDE``
    object. To plot it, use ``pymde.plot``.

    .. code:: python3

        embedding = pymde.preserve_neighbors(data).embed()
        pymde.plot(embedding)

    Arguments
    ---------
    data: {torch.Tensor, numpy.ndarray, scipy.sparse matrix} or pymde.Graph
        The original data, a data matrix of shape ``(n_items, n_features)`` or
        a graph. Neighbors are computed using Euclidean distance if the data is
        a matrix, or the shortest-path metric if the data is a graph.
    embedding_dim: int
        The embedding dimension. Use 2 or 3 for visualization.
    attractive_penalty: pymde.Function class (or factory)
        Callable that constructs a distortion function, given positive
        weights. Typically one of the classes from ``pymde.penalties``,
        such as ``pymde.penalties.log1p``, ``pymde.penalties.Huber``, or
        ``pymde.penalties.Quadratic``.
    repulsive_penalty: pymde.Function class (or factory)
        Callable that constructs a distortion function, given negative
        weights. (If ``None``, only positive weights are used.) For example,
        ``pymde.penalties.Log`` or ``pymde.penalties.InversePower``.
    constraint: pymde.constraints.Constraint (optional)
        Embedding constraint, like ``pymde.Standardized()`` or
        ``pymde.Anchored(anchors, values)`` (or ``None``). Defaults to no
        constraint when a repulsive penalty is provided, otherwise defaults to
        ``pymde.Standardized()``.
    n_neighbors: int (optional)
        The number of nearest neighbors to compute for each row (item) of
        ``data``. A sensible value is chosen by default, depending on the
        number of items.
    repulsive_fraction: float (optional)
        How many repulsive edges to include, relative to the number
        of attractive edges. ``1`` means as many repulsive edges as attractive
        edges. The higher this number, the more uniformly spread out the
        embedding will be. Defaults to ``0.5`` for standardized embeddings, and
        ``1`` otherwise. (If ``repulsive_penalty`` is ``None``, this argument
        is ignored.)
    max_distance: float (optional)
        If not None, neighborhoods are restricted to have a radius
        no greater than ``max_distance``.
    init: str
        Initialization strategy; 'quadratic' or 'random'. If the quadratic
        initialization takes too much time, try a random initialization.
        When `device=='cuda'` and `init=='quadratic'` a torch-implemented
        version will be used, which is fully on device, but uses a CG-based
        technique (versus Lanczos).
    device: str (optional)
        Device for the embedding (eg, 'cpu', 'cuda').
    verbose: bool
        If ``True``, print verbose output.

    Returns
    -------
    pymde.MDE
        A ``pymde.MDE`` object, based on the original data.
    """
    if isinstance(data, preprocess.graph.Graph):
        n = data.n_items
    elif data.shape[0] <= 1:
        raise ValueError("The data matrix must have at least two rows.")
    else:
        n = data.shape[0]

    if n_neighbors is None:
        # target included edges to be ~1% of total number of edges
        n_choose_2 = n * (n - 1) / 2
        n_neighbors = int(max(min(15, n_choose_2 * 0.01 / n), 5))
    if n_neighbors > n:
        problem.LOGGER.warning(
            (
                "Requested n_neighbors {0} > number of items {1}."
                " Setting n_neighbors to {2}"
            ).format(n_neighbors, n, n - 1)
        )
        n_neighbors = n - 1

    if constraint is None and repulsive_penalty is not None:
        constraint = constraints.Centered()
    elif constraint is None and repulsive_penalty is None:
        constraint = constraints.Standardized()

    if isinstance(data, preprocess.graph.Graph):
        # enforce a max distance, otherwise may very well run out of memory
        # when n_items is large
        if max_distance is None:
            max_distance = (3 * torch.quantile(data.distances, 0.75)).item()

    if verbose:
        problem.LOGGER.info(
            f"Computing {n_neighbors}-nearest neighbors, with "
            f"max_distance={max_distance}"
        )

    knn_graph = preprocess.generic.k_nearest_neighbors(
        data,
        k=n_neighbors,
        max_distance=max_distance,
        verbose=verbose,
    )
    edges = knn_graph.edges.to(device)
    weights = knn_graph.weights.to(device)

    if isinstance(constraint, constraints.Anchored):
        # remove anchor-anchor edges before generating intialization
        edges, weights = _remove_anchor_anchor_edges(
            edges, weights, constraint.anchors
        )

    if init == "quadratic":
        if verbose:
            problem.LOGGER.info(f"Computing {init} initialization.")
        # use cg + torch when using GPU
        cg = device == "cuda"
        X_init = quadratic.spectral(
            n,
            embedding_dim,
            edges,
            weights,
            max_iter=1000,
            device=device,
            cg=cg,
        )
        if not isinstance(
            constraint, (constraints._Centered, constraints._Standardized)
        ):
            constraint.project_onto_constraint(X_init, inplace=True)
    elif init == "random":
        X_init = constraint.initialization(n, embedding_dim, device)
    else:
        raise ValueError(
            f"Unsupported value '{init}' for keyword argument `init`; "
            "the supported values are 'quadratic' and 'random'."
        )

    if repulsive_penalty is not None:
        if repulsive_fraction is None:
            if isinstance(constraint, constraints._Standardized):
                # standardization constraint already implicity spreads,
                # so use a lower replusion
                repulsive_fraction = 0.5
            else:
                repulsive_fraction = 1

        n_choose_2 = int(n * (n - 1) / 2)
        n_repulsive = int(repulsive_fraction * (edges.shape[0]))
        # cannot sample more edges than there are available
        n_repulsive = min(n_repulsive, n_choose_2 - edges.shape[0])

        negative_edges = preprocess.sample_edges(
            n, n_repulsive, exclude=edges
        ).to(device)

        negative_weights = -torch.ones(
            negative_edges.shape[0], dtype=X_init.dtype, device=device
        )

        if isinstance(constraint, constraints.Anchored):
            negative_edges, negative_weights = _remove_anchor_anchor_edges(
                negative_edges, negative_weights, constraint.anchors
            )

        edges = torch.cat([edges, negative_edges])
        weights = torch.cat([weights, negative_weights])

        f = penalties.PushAndPull(
            weights,
            attractive_penalty=attractive_penalty,
            repulsive_penalty=repulsive_penalty,
        )
    else:
        f = attractive_penalty(weights)

    mde = problem.MDE(
        n_items=n,
        embedding_dim=embedding_dim,
        edges=edges,
        distortion_function=f,
        constraint=constraint,
        device=device,
    )
    mde._X_init = X_init

    # TODO cache the graph for subsequent calls / constructor for MDE from graph

    distances = mde.distances(mde._X_init)
    if (distances == 0).any():
        # pathological scenario in which at least two points overlap can yield
        # non-differentiable average distortion. perturb the initialization to
        # mitigate.
        mde._X_init += 1e-4 * torch.randn(
            mde._X_init.shape,
            device=mde._X_init.device,
            dtype=mde._X_init.dtype,
        )
    return mde


def laplacian_embedding(
    data,
    embedding_dim=2,
    n_neighbors=None,
    max_distance=None,
    init="quadratic",
    device="cpu",
    verbose=False,
) -> problem.MDE:
    """Constructs an MDE problem for computing a Laplacian embedding.

    The embedding preserves the nearest neighbors of each item, using
    quadratic distortion functions and a standardization constraint.

    Arguments
    ---------
    data: {torch.Tensor, numpy.ndarray, scipy.sparse matrix} or pymde.Graph
        The original data, a data matrix of shape ``(n_items, n_features)`` or
        a graph. Neighbors are computed using Euclidean distance if the data is
        a matrix, or the shortest-path metric if the data is a graph.
    embedding_dim: int
        The embedding dimension. Use 2 or 3 for visualization.
    n_neighbors: int (optional)
        The number of nearest neighbors to compute for each row (item) of
        ``data``. A sensible value is chosen by default, depending on the
        number of items.
    max_distance: float (optional)
        If not None, neighborhoods are restricted to have a radius
        no greater than ``max_distance``.
    init: str
        Initialization strategy; 'quadratic' or 'random'. If the quadratic
        initialization takes too much time, try a random initialization.
    device: str (optional)
        Device for the embedding (eg, 'cpu', 'cuda').
    verbose: bool
        If ``True``, print verbose output.

    Returns
    -------
    pymde.MDE
        A ``pymde.MDE`` object, based on the original data.
    """
    return preserve_neighbors(
        data,
        embedding_dim=embedding_dim,
        attractive_penalty=penalties.Quadratic,
        repulsive_penalty=None,
        n_neighbors=n_neighbors,
        max_distance=max_distance,
        init=init,
        device=device,
        verbose=verbose,
    )



import logging


import torch


LOGGER = logging.getLogger("__pymde__")


# PyTorch has an inneficient implementation of the 2-norm; this
# implementation is about 2x faster on a GPU.
class _Norm(torch.autograd.Function):
    @staticmethod
    def forward(ctx, x):
        norm_x = x.pow(2).sum(dim=1).sqrt()
        ctx.save_for_backward(x, norm_x)
        return norm_x

    @staticmethod
    def backward(ctx, grad_output):
        x, norm_x = ctx.saved_tensors
        grad_output = grad_output.unsqueeze(1)
        norm_x = norm_x.unsqueeze(1)
        grad_input = x.mul(grad_output).div(norm_x)
        grad_input[torch.isnan(grad_input)] = 0.0
        return grad_input


_norm = _Norm.apply


def _gather_indices(idx, m):
    return idx[:, None].expand(idx.shape[0], m)


class _AverageDistortion(torch.autograd.Function):
    @staticmethod
    def forward(ctx, X, f, lhs, rhs):
        ctx.shape = X.shape

        # indexing appears to be faster on cpu and cuda
        differences = X[lhs[:, 0]] - X[rhs[:, 0]]
        # differences.pow(2) allocates another p*m*4 bytes,
        # which can OOM a GPU when p is large, so square it in place and
        # sqrt root it when not needed
        norms = differences.pow(2).sum(dim=1).sqrt()
        if X.requires_grad:
            with torch.enable_grad():
                norms.requires_grad_(True)
                norms.grad = None
                distortion = f(norms).mean()
                distortion.backward()
                norms.requires_grad_(False)
            # norms can be zero (rarely)
            g = norms.grad / norms
            # TODO(akshayka): Log a warning on first occurrence during solve.
            if (norms == 0).any():
                pass
            if torch.isnan(g).any():
                g[torch.isnan(g)] = 1.0
            if torch.isinf(g).any():
                g[torch.isinf(g)] = 1.0
            ctx.save_for_backward(differences, g, lhs, rhs)
        else:
            distortion = f(norms).mean()
        return distortion

    @staticmethod
    def backward(ctx, grad_output):
        out = torch.zeros(
            ctx.shape, dtype=grad_output.dtype, device=grad_output.device
        )
        differences, g, lhs, rhs = ctx.saved_tensors
        DATX = g[:, None] * differences
        # scatter add is faster than index_put on cuda, and on CPU
        # (index_put is serial on CPU) ... but scatter_add is nondeterministic
        out.scatter_add_(0, lhs, DATX)
        out.scatter_add_(0, rhs, -DATX)
        out.mul_(grad_output)
        return out, None, None, None


_average_distortion = _AverageDistortion.apply


class _ProjectGradient(torch.autograd.Function):
    @staticmethod
    def forward(ctx, X, constraint):
        ctx.constraint = constraint
        ctx.X = X
        return X

    @staticmethod
    def backward(ctx, grad_output):
        return (
            ctx.constraint.project_onto_tangent_space(
                ctx.X, grad_output, inplace=False
            ),
            None,
        )


_project_gradient = _ProjectGradient.apply



"""Minimum-Distortion Embedding

This module defines the MDE class, which represents an MDE problem.
"""


import copy
import glob
import logging
import sys
import tempfile
import typing as tp

import torch

from pymde.average_distortion import _average_distortion
from pymde.average_distortion import _gather_indices, _norm
from pymde import constraints
from pymde.functions.function import StochasticFunction
from pymde.util import _canonical_device, _module_device
from pymde import optim


LOGGER = logging.getLogger("__pymde__")
LOGGER.propagate = False
LOGGER.setLevel(logging.INFO)
_stream_handler = logging.StreamHandler(sys.stdout)
_stream_handler.setLevel(logging.INFO)
_formatter = logging.Formatter(
    fmt="%(asctime)s: %(message)s", datefmt="%b %d %I:%M:%S %p"
)
_stream_handler.setFormatter(_formatter)
LOGGER.addHandler(_stream_handler)


class MDE(torch.nn.Module):
    """An MDE problem.

    An MDE instance represents a specific MDE problem, specified by
    the number of items, the embedding dimension, a list of edges,
    the vector distortion function, and possibly a constraint.

    Attributes
    ----------
    n_items: int
        The number of items
    embedding_dim: int
        The embedding dimension
    edges: torch.Tensor
        The list of edges.
    distortion_function: Callable
        The vector distorton function.
    constraint: pymde.constraints.Constraint
        The constraint imposed (or None)
    device
        The device on which to compute the embedding and store data
        (like 'cpu' or 'cuda')
    solve_stats
        Summary statistics about the embedding, populated after calling
        the ``embed`` method.
    """

    def __init__(
        self,
        n_items: int,
        embedding_dim: int,
        edges: torch.Tensor,
        distortion_function: tp.Union[tp.Callable, StochasticFunction],
        constraint: tp.Optional[constraints.Constraint] = None,
        device: tp.Optional[str] = None,
    ):
        """Constructs an MDE problem.

        Arguments
        ---------
        n_items: int
            Number of things being embedded.
        embedding_dim: int
            Embedding dimension.
        edges: torch.Tensor(shape=(num_edges, 2), dtype=torch.int)
            Tensor, where each row is an edge (i, j) between two items;
            each edge should satisfy 0 <= i < j < n_items. In particular
            self-edges are not allowed.
        distortion_function: Callable or pymde.functions.StochasticFunction
            The vectorized distortion function, typically an instance
            of a class from ``pymde.penalties`` or ``pymde.losses`` however,
            this can be any Python callable that maps a torch.Tensor
            of embedding distances to a torch.Tensor of distortions.
        constraint: pymde.constraints.Constraint, optional
            A Constraint object, such as ``pymde.Standardized()``
            Defaults to an unconstrained (centered) embedding.
        device: str, optional
            Name of device on which to store tensors/compute embedding,
            such as 'cpu' or 'cuda' for GPU. Default infers device from
            ``edges`` and ``distortion_function``
        """
        super(MDE, self).__init__()
        if device is None:
            if (
                isinstance(edges, torch.Tensor)
                and isinstance(distortion_function, torch.nn.Module)
            ) and (
                str(edges.device) == str(_module_device(distortion_function))
            ):
                device = edges.device
            else:
                device = "cpu"
        self.device = _canonical_device(device)

        if not isinstance(n_items, torch.Tensor):
            n_items = torch.tensor(n_items, device=self.device)
        elif str(n_items.device) != str(self.device):
            n_items = n_items.to(self.device)
        self.register_buffer("n_items", n_items)

        if not isinstance(embedding_dim, torch.Tensor):
            embedding_dim = torch.tensor(embedding_dim, device=self.device)
        elif str(embedding_dim.device) != str(self.device):
            embedding_dim = embedding_dim.to(self.device)
        self.register_buffer("embedding_dim", embedding_dim)

        if edges is None:
            if not isinstance(distortion_function, StochasticFunction):
                raise ValueError(
                    "edges can only be None when using a stochastic function."
                )
            p = distortion_function.p
        else:
            if not isinstance(edges, torch.Tensor):
                edges = torch.tensor(
                    edges, dtype=torch.int64, device=self.device
                )

            if (edges[:, 0] == edges[:, 1]).any():
                offending = torch.where(edges[:, 0] == edges[:, 1])[0]
                raise ValueError(
                    "The edge list must not contain self edges; the "
                    "following rows were found to be self edges: ",
                    offending.cpu().numpy(),
                )

            if str(edges.device) != str(self.device):
                LOGGER.warning(
                    "edges.device (%s) "
                    "does not match requested device (%s); copying edges to "
                    "requested device." % (edges.device, device)
                )
                edges = edges.to(self.device)
            p = torch.tensor(edges.shape[0], device=self.device)

        complete_graph_edges = (n_items * (n_items - 1) / 2).long()
        if p is not None and p > complete_graph_edges:
            raise ValueError(
                "Your graph has more than (n_items choose 2) edges."
                "(p: {0}, n_items choose 2: {1})".format(
                    p, complete_graph_edges
                )
            )

        self.register_buffer("edges", edges)
        self.register_buffer("p", p)
        self.register_buffer("_complete_graph_edges", complete_graph_edges)

        if edges is not None:
            self.register_buffer(
                "_lhs", _gather_indices(edges[:, 0], self.embedding_dim)
            )
            self.register_buffer(
                "_rhs", _gather_indices(edges[:, 1], self.embedding_dim)
            )

        if isinstance(distortion_function, torch.nn.Module):
            f_device = _module_device(distortion_function)
            if f_device is None or str(f_device) != str(self.device):
                LOGGER.warning(
                    "distortion_function device (%s) "
                    "does not match requested device (%s); making a copy of "
                    "distortion_function" % (str(f_device), device)
                )
                distortion_function = copy.deepcopy(distortion_function)
                distortion_function.to(self.device)
        self.distortion_function = distortion_function

        if constraint is None:
            constraint = constraints.Centered()
        self.constraint = constraint

        self.register_buffer("X", None)
        self.register_buffer("_X_init", None)

        self.solve_stats = None
        self.value = None
        self.residual_norm = None

    def to(self, device):
        """Move MDE instance to another device."""
        super(MDE, self).to(device)
        self.device = _canonical_device(device)

    def __str__(self):
        func_name = (
            self.distortion_function.__name__
            if hasattr(self.distortion_function, "__name__")
            else type(self.distortion_function).__name__
        )
        if self.p is None:
            return (
                "Stochastic MDE problem:\n"
                "\tn (number of items) {0}\n"
                "\tm (embedding dimension) {1}\n"
                "\t{2} distortion functions\n"
                "\tconstraint {3}\n"
                "\tdevice {4}".format(
                    self.n_items.item(),
                    self.embedding_dim.item(),
                    func_name,
                    self.constraint.name(),
                    self.device,
                )
            )

        return (
            "MDE problem:\n"
            "\tn (number of items) {0}\n"
            "\tm (embedding dimension) {1}\n"
            "\tp (number of edges) {2}\n"
            "\tfraction of total edges {3:.1e}\n"
            "\t{4} distortion functions\n"
            "\tconstraint {5}\n"
            "\tdevice {6}".format(
                self.n_items.item(),
                self.embedding_dim.item(),
                self.p,
                (float(self.p) / self._complete_graph_edges).item(),
                func_name,
                self.constraint.name(),
                self.device,
            )
        )

    def _repr_pretty_(self, p, cycle):
        del cycle
        text = self.__str__()
        p.text(text)

    def differences(self, X):
        """Compute ``X[i] - X[j]`` for each row (i, j) in edges."""
        lhs = X.gather(0, self._lhs)
        rhs = X.gather(0, self._rhs)
        return lhs.sub_(rhs)

    def distances(self, X=None):
        """Compute embedding distances.

        This function computes the embedding distances corresponding to
        the list of edges (``self.edges``)

        If ``X`` is None, computes distances for the embedding computed
        by the last call to ``embed``

        Arguments
        ---------
        X: torch.Tensor(shape=(n_items, 2)), optional
            Embedding.

        Returns
        -------
        torch.Tensor(shape=(n_edges,))
            Vector of embedding distances. The k-th entry is the distance
            corresponding to the k-th edge (``self.edges[k]``
        """
        if X is None:
            X = self.X
        if X is None:
            raise ValueError(
                "Call this function after running the `embed` method, or "
                "provide a value for the embedding argument `X`"
            )
        return _norm(self.differences(X))

    def distortions(self, X=None):
        """Compute distortions

        This function computes the distortions for an embedding.

        If ``X`` is None, computes distortions for the embedding computed
        by the last call to ``embed``

        Arguments
        ---------
        X: torch.Tensor(shape=(n_items, 2)), optional
            Embedding.

        Returns
        -------
        torch.Tensor(shape=(n_edges,))
            Vector of distortions. The k-th entry is the distortion
            for the k-th edge (``self.edges[k]``
        """
        if X is None:
            X = self.X
        if X is None:
            raise ValueError(
                "Call this function after running the `embed` method, or "
                "provide a value for the embedding argument `X`"
            )
        return self.distortion_function(self.distances(X))

    def average_distortion(self, X=None):
        """Compute average distortion.

        This method computes the average distortion of an embedding.

        If ``X`` is None, this method computes the average distortion for the
        embedding computed by the last call to ``embed``.

        Arguments
        ---------
        X: torch.Tensor(shape=(n_items, 2)), optional
            Embedding.

        Returns
        -------
        scalar (float)
            The average distortion.
        """
        if X is None:
            X = self.X
        if X is None:
            raise ValueError(
                "Call this function after running the `embed` method, or "
                "provide a value for the embedding argument `X`"
            )
        return _average_distortion(
            X, self.distortion_function, self._lhs, self._rhs
        )

    def high_distortion_pairs(self, X=None):
        """Compute distortions, sorted from high to low.

        Computes the distortions for an embedding, sorting them from high
        to low. Returns the edges and distortions in the sorted order.

        This function can be used to debug an embedding or search for outliers
        in the data. In particular it can be instructive to manually
        examine the items which incurred the highest distortion.

        For example:

        .. code:: python3

            edges, distortions = mde.high_distortion_pairs()
            maybe_outliers = edges[:10]

        You can then examine the items corresponding to the edges in
        ``maybe_outliers``.

        Arguments
        ---------
        X: torch.Tensor(shape=(n_items, 2)), optional
            Embedding.

        Returns
        -------
        torch.Tensor(shape=(n_edges, 2))
            edges, sorted from highest distortion to lowest

        torch.Tensor(shape=(n_edges,))
            distortions, sorted from highest to lowest
        """
        if X is None:
            X = self.X

        if X is None:
            raise ValueError(
                "Call this function after running the `embed` method, or "
                "provide a value for the embedding argument `X`"
            )
        distortions = self.distortions(X)
        # everything after `argsort()` just reverses the array
        indices = distortions.argsort()[:, None].flipud().flatten()
        distortions_high_to_low = distortions[indices]
        sorted_pairs = self.edges[indices]
        return sorted_pairs, distortions_high_to_low

    def embed(
        self,
        X=None,
        eps=1e-5,
        max_iter=300,
        memory_size=10,
        verbose=False,
        print_every=None,
        snapshot_every=None,
    ):
        """Compute an embedding.

        This method stores the embedding in the ``X`` attribute of the problem
        instance (``mde.X``). Summary statistics related to the fitting
        process are stored in ``solve_stats`` (``mde.solve_stats``).

        All arguments have sensible default values, so in most cases,
        it suffices to just type ``mde.embed()`` or ``mde.embed(verbose=True)``

        Arguments
        ---------
        X: torch.Tensor, optional
            Initial iterate, of shape ``(n_items, embedding_dim)``. When None,
            the initial iterate is chosen randomly (and projected onto the
            constraints); otherwise, the initial iterate should satisfy
            the constraints.
        eps: float
            Residual norm threshold; quit when the residual norm
            is smaller than ``eps``.
        max_iter: int
            Maximum number of iterations.
        memory_size: int
            The quasi-Newton memory. Larger values may lead to more stable
            behavior, but will increase the amount of time each iteration
            takes.
        verbose: bool
            Whether to print verbose output.
        print_every: int, optional
            Print verbose output every ``print_every`` iterations.
        snapshot_every: int, optional
            Snapshot embedding every ``snapshot_every`` iterations;
            snapshots saved as CPU tensors to ``self.solve_stats.snapshots``.
            If you want to generate an animation with the ``play`` method after
            embedding, set ``snapshot_every`` to a positive integer
            (like 1 or 5).

        Returns
        -------
        torch.Tensor
            The embedding, of shape ``(n_items, embedding_dim)``.
        """
        if isinstance(self.distortion_function, StochasticFunction):
            device = (self.distortion_function.device,)
        else:
            device = self.device

        if X is None and self._X_init is not None:
            X = self._X_init.detach().clone()
        elif X is None:
            X = self.constraint.initialization(
                self.n_items, self.embedding_dim, device
            )
        else:
            X = X.detach().clone()

        if X.device != self.device:
            LOGGER.warning(
                f"The initial iterate's device ({X.device}) does not match "
                f"the requested device ({device}). Copying the iterate to "
                f"{device}."
            )
            X = X.to(device)

        if max_iter < 0:
            raise ValueError("`max_iter` must be greater than 0")
        if memory_size <= 0:
            raise ValueError("`memory_size` must be greater than 0")
        requires_grad = X.requires_grad

        if verbose:
            LOGGER.info(
                (
                    f"Fitting a {self.constraint.name()} embedding into "
                    f"R^{int(self.embedding_dim)}, for a graph with "
                    f"{int(self.n_items)} items and {int(self.p)} edges."
                )
            )
            LOGGER.info(
                f"`embed` method parameters: eps={eps:.1e}, "
                f"max_iter={max_iter}, memory_size={memory_size}"
            )

        if print_every is None:
            print_every = max(1, max_iter // 10)

        if isinstance(self.distortion_function, StochasticFunction):
            X_star, solve_stats = optim.spi(
                batch_size=self.distortion_function.batch_size,
                stochastic_function=self.distortion_function,
                X=X,
                constraint=self.constraint,
                eps=eps,
                max_iter=max_iter,
                memory_size=memory_size,
                use_line_search=True,
                verbose=verbose,
                print_every=print_every,
                snapshot_every=snapshot_every,
                logger=LOGGER,
            )
        else:
            X_star, solve_stats = optim.lbfgs(
                X=X,
                constraint=self.constraint,
                objective_fn=self.average_distortion,
                eps=eps,
                max_iter=max_iter,
                memory_size=memory_size,
                use_line_search=True,
                use_cached_loss=True,
                verbose=verbose,
                print_every=print_every,
                snapshot_every=snapshot_every,
                logger=LOGGER,
            )

        self.X = X_star
        self.solve_stats = solve_stats
        self.value = solve_stats.average_distortions[-1]
        self.residual_norm = solve_stats.residual_norms[-1]

        if verbose:
            LOGGER.info(
                f"Finished fitting in {solve_stats.solve_time:.3f} seconds "
                f"and {solve_stats.iterations} iterations."
            )
            LOGGER.info(
                f"average distortion {self.value:.3g} | "
                f"residual norm {self.residual_norm:.1e}"
            )
        X.requires_grad_(requires_grad)
        return self.X

    forward = embed

    def plot(
        self,
        color_by=None,
        color_map="Spectral",
        colors=None,
        edges=None,
        axis_limits=None,
        background_color=None,
        marker_size=1.0,
        figsize_inches=(8.0, 8.0),
        savepath=None,
    ):
        """Plot an embedding, in one, two, or three dimensions.

        This method plots embeddings, with embedding dimension at most 3.

        The embedding is visualized as a scatter plot. The points can
        optionally be colored according to categorical or continuous values,
        or according to a pre-defined sequence of colors. Additionally,
        edges can optionally be superimposed.

        The ``embed`` method must be called sometime before calling this method.

        Arguments
        ---------
        color_by: np.ndarray(shape=mde.n_items), optional
            A sequence of values, one for each item, which should be
            used to color each embedding vector. These values may either
            be categorical or continuous. For example, if ``n_items`` is 4,

            .. code:: python3

                np.ndarray(['dog', 'cat', 'zebra', 'cat'])
                np.ndarray([0, 1, 1, 2]
                np.ndarray([0.1, 0.5, 0.31, 0.99]

            are all acceptable. The first two are treated as categorical,
            the third is continuous. A finite number of colors is used
            when the values are categorical, while a spectrum of colors is
            used when the values are continuous.
        color_map: str or matplotlib colormap instance
            Color map to use when resolving ``color_by`` to colors; ignored
            when ``color_by`` is None.
        colors: np.ndarray(shape=(mde.n_items, 4)), optional
            A sequence of colors, one for each item, specifying the exact
            color each item should be colored. Each row must represent
            an RGBA value.

            Only one of ``color_by`` and ``colors`` should be non-None.
        edges: {torch.Tensor/np.ndarray}(shape=(any, 2)), optional
            List of edges to superimpose over the scatter plot.
        axis_limits: tuple(length=2), optional
            tuple (limit_low, limit_high) of axis limits, applied to both
            the x and y axis.
        background_color: str, optional
            color of background
        marker_size: float, optional
            size of each point in the scatter plot
        figsize_inches: tuple(length=2)
            size of figures in inches: (width_inches, height_inches)
        savepath: str, optional
            path to save the plot.

        Returns
        -------
        matplotlib axis:
            Axis on which the embedding is plotted.
        """
        from pymde import experiment_utils

        if self.X is None:
            raise ValueError("The `embed` method must be called before `show`")

        if color_by is not None and colors is not None:
            raise ValueError(
                "Only one of `color_by` and `colors` "
                "should be provided, not both."
            )

        return experiment_utils.plot(
            X=self.X,
            color_by=color_by,
            color_map=color_map,
            colors=colors,
            edges=edges,
            axis_limits=axis_limits,
            background_color=background_color,
            figsize_inches=figsize_inches,
            marker_size=marker_size,
            savepath=savepath,
        )

    def distortions_cdf(self, figsize_inches=(8, 3)):
        """Plot a cumulative distribution function of the distortions.

        The ``embed`` method must be called sometime before calling this method.
        """
        from pymde import experiment_utils

        if self.X is None:
            raise ValueError(
                "This method can only be called after calling `embed`"
            )
        experiment_utils._show_cdf(self, figsize_inches=figsize_inches)

    def play(
        self,
        color_by=None,
        color_map="Spectral",
        colors=None,
        edges=None,
        axis_limits=None,
        background_color=None,
        marker_size=1.0,
        figsize_inches=(8.0, 8.0),
        fps=None,
        tmpdir=None,
        savepath=None,
    ):
        """Create a movie visualizing how the embedding was formed.

        This method creates a GIF showing how the embedding was formed, starting
        with the initial iterate and ending with the final embedding. In other
        words it visualizes how the embedding algorithm (the ``embed`` method)
        updated the embedding's state over time. (The embedding dimension should
        be at most 3.)

        In each frame, the embedding is visualized as a scatter plot. The points
        can optionally be colored according to categorical or continuous
        values, or according to a pre-defined sequence of colors. Additionally,
        edges can optionally be superimposed.

        The ``embed`` method must be called sometime before calling this method,
        with a non-None value for the ``snapshot_every`` keyword argument.

        If you want to save the GIF (instead of just viewing it in a Jupyter
        notebook), make sure to supply a path via the ``savepath`` keyword
        argument.

        Arguments
        ---------
        color_by: np.ndarray(shape=mde.n_items), optional
            A sequence of values, one for each item, which should be
            used to color each embedding vector. These values may either
            be categorical or continuous. For example, if ``n_items`` is 4,

            .. code:: python3

                categorical_values = np.ndarray(['dog', 'cat', 'zebra', 'cat'])
                other_categorical_values = np.ndarray([0, 1, 1, 2]
                continuous_values = np.ndarray([0.1, 0.5, 0.31, 0.99]

            are all acceptable values for this argument. A finite number of
            colors is used when the values are categorical, while a
            spectrum of colors is used when the values are continuous.
        color_map: str or matplotlib colormap instance
            Color map to use when resolving ``color_by`` to colors; ignored
            when ``color_by`` is None.
        colors: np.ndarray(shape=(mde.n_items, 4)), optional
            A sequence of colors, one for each item, specifying the exact
            color each item should be colored. Each row must represent
            an RGBA value.

            Only one of ``color_by`` and ``colors`` should be non-None.
        edges: {torch.Tensor/np.ndarray}(shape=(any, 2)), optional
            List of edges to superimpose over the scatter plot.
        axis_limits: tuple(length=2), optional
            tuple (limit_low, limit_high) of axis limits, applied to both
            the x and y axis.
        background_color: str, optional
            color of background
        marker_size: float, optional
            size of each point in the scatter plot
        figsize_inches: tuple(length=2)
            size of figures in inches: (width_inches, height_inches)
        fps : float, optional
            the number of frames per second at which the movie should be
            shown
        tmpdir: str, optional
            path to directory, where individual images comprising the GIF
            will be stored
        savepath: str, optional
            path at which to save the GIF.
        """
        if self.solve_stats is None or not self.solve_stats.snapshots:
            raise ValueError(
                "The play method can only be called after "
                "calling embed with the snapshot_every keyword argument, "
                "as in mde.embed(snapshot_every=5)"
            )

        from pymde import experiment_utils

        try:
            import PIL
        except ImportError:
            raise ImportError("Install PIL to use this method.")

        try:
            from tqdm.auto import tqdm
        except ImportError:
            tqdm = lambda x: x

        if tmpdir is None:
            tmpdir_obj = tempfile.TemporaryDirectory()
            tmpdir = tmpdir_obj.name

        if axis_limits is None:
            lim_low = torch.tensor(float("inf"))
            lim_high = -torch.tensor(float("inf"))
            for shot in self.solve_stats.snapshots:
                low = shot.min()
                high = shot.max()
                if low < lim_low:
                    lim_low = low
                if high > lim_high:
                    lim_high = high
            axis_limits = experiment_utils._square_lim(lim_low, lim_high)

        for i in tqdm(range(len(self.solve_stats.snapshots) + 1)):
            # Repeat the last frame just once, to make the final embedding
            # look sticky.
            #
            # TODO: doing this in a multiprocessing map hangs when run under
            # a jupyter notebook, and disabling interactive mode (plt.ioff())
            # doesn't fix it
            experiment_utils._plot_gif_frame(
                self.solve_stats.snapshots[
                    min(len(self.solve_stats.snapshots) - 1, i)
                ],
                color_by=color_by,
                cmap=color_map,
                colors=colors,
                s=marker_size,
                lim=axis_limits,
                background_color=background_color,
                title=None,
                edges=edges,
                figsize_inches=figsize_inches,
                i=i,
                outdir=tmpdir,
            )

        if fps is None:
            # aim for a 4 second movie, but no fewer than 15 fps
            n_frames = len(self.solve_stats.snapshots)
            fps = max(n_frames / 4.0, 15.0)

        frame_duration = 1000.0 / fps
        fp_in = tmpdir + "/*.png"
        if savepath is None:
            fp_out = tmpdir + "/mde.gif"
        else:
            fp_out = savepath

        img, *imgs = [PIL.Image.open(f) for f in sorted(glob.glob(fp_in))]
        img.save(
            fp=fp_out,
            format="GIF",
            append_images=imgs,
            save_all=True,
            duration=frame_duration,
            loop=0,
        )

        try:
            from IPython.display import Image, display

            with open(fp_out, "rb") as f:
                display(Image(f.read()))
        except ImportError:
            pass



from contextlib import contextmanager
import functools
import logging

import numpy as np
import pytest
import torch

from pymde import util


def assert_allclose(x, y, up_to_sign=False, rtol=1e-4, atol=1e-5):
    if isinstance(x, torch.Tensor):
        x = x.detach().cpu().numpy()
    if isinstance(y, torch.Tensor):
        y = y.detach().cpu().numpy()
    if up_to_sign:
        try:
            np.testing.assert_allclose(x, y, rtol=rtol, atol=atol)
        except AssertionError:
            np.testing.assert_allclose(-x, y, rtol=rtol, atol=atol)
    else:
        np.testing.assert_allclose(x, y, rtol=rtol, atol=atol)


def assert_all_equal(x, y):
    if isinstance(x, torch.Tensor):
        x = x.detach().cpu().numpy()
    if isinstance(y, torch.Tensor):
        y = y.detach().cpu().numpy()
    np.testing.assert_array_equal(x, y)


def cpu(func):
    @functools.wraps(func)
    def wrapper(self):
        current_device = util.get_default_device()
        util.set_default_device("cpu")
        func(self)
        util.set_default_device(current_device)

    return wrapper


def cpu_and_cuda(func):
    if torch.cuda.is_available():
        return pytest.mark.parametrize("device", ["cpu", "cuda"])(func)
    else:
        return pytest.mark.parametrize("device", ["cpu"])(func)


@contextmanager
def disable_logging(up_to=logging.CRITICAL):
    previous_level = logging.root.manager.disable

    logging.disable(up_to)

    try:
        yield
    finally:
        logging.disable(previous_level)



import abc

import torch
from pymde import util


class Constraint(abc.ABC):
    """A generic constraint.

    To create a custom constraint, create a subclass of this class,
    and implement its abstract methods.
    """

    @abc.abstractmethod
    def name(self) -> str:
        """The name of the constraint."""
        raise NotImplementedError

    @abc.abstractmethod
    def initialization(
        self, n_items: int, embedding_dim: int, device=None
    ) -> torch.Tensor:
        """Return a random embedding in the constraint set.

        Arguments
        ---------
        n_items: int
            The number of items.
        embedding_dim: int
            The embedding dimension.
        device: str
            Device on which to store the returned embedding.

        Returns
        -------
        torch.Tensor
            a tensor of shape ``(n_items, embedding_dim)`` satisfying the
            constraints.
        """

        raise NotImplementedError

    @abc.abstractmethod
    def project_onto_constraint(
        self, Z: torch.Tensor, inplace=True
    ) -> torch.Tensor:
        """Project ``Z`` onto the constraint set.

        Returns a projection of ``Z`` onto the constraint set.


        Arguments
        ---------
        Z: torch.Tensor
            The point to project.
        inplace: bool
            If True, stores the projection in ``Z``.

        Returns
        -------
        The projection of ``Z`` onto the constraints.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def project_onto_tangent_space(
        self, X: torch.Tensor, Z: torch.Tensor, inplace=True
    ) -> torch.Tensor:
        """Project ``Z`` onto the tangent space of the constraint set at ``X``.

        Returns the Euclidean projection of ``Z`` onto
        the tangent space of the constraint set at ``X`` (where ``X`` is
        some matrix satisfying the constraints).

        ``X`` and ``Z`` should have the same shape.

        Arguments
        ---------
        X: torch.Tensor
            A point satisfying the constraints.
        Z: torch.Tensor
            The point to project.
        inplace: bool
            If True, stores the projection in ``Z``.

        Return
        ------
        The projection of ``Z`` onto the tangent space of the constraint
        set at ``X``.
        """
        raise NotImplementedError


class _Centered(Constraint):
    def name(self):
        return "centered"

    def initialization(self, n_items, embedding_dim, device=None):
        X = torch.randn((n_items, embedding_dim), device=device)
        return X - X.mean(axis=0)

    def project_onto_tangent_space(self, X, Z, inplace=True):
        del X
        return Z

    def project_onto_constraint(self, Z, inplace=True):
        if inplace:
            Z.sub_(Z.mean(axis=0))
            return Z
        else:
            return Z - Z.mean(axis=0)


class Anchored(Constraint):
    """Anchor some vectors to specific values."""

    def __init__(self, anchors, values):
        """
        Constructs an anchor constraint, in which some embedding vectors
        (the anchors) are fixed to specific values.

        Arguments
        ---------
        anchors: torch.Tensor, shape (n_anchors)
            a Tensor in which each entry gives the index of an anchored
            vertex
        values: torch.Tensor, shape (n_anchors, embedding_dim)
            a Tensor which gives the value to which each anchor should be
            fixed
        """
        super(Anchored, self).__init__()
        self.anchors = anchors
        self.values = values

    def name(self):
        return "anchored"

    def initialization(self, n_items, embedding_dim, device=None):
        X = torch.randn((n_items, embedding_dim), device=device)
        X[self.anchors] = self.values
        return X

    def project_onto_tangent_space(self, X, Z, inplace=True):
        del X
        if inplace:
            Z[self.anchors, :] = 0.0
        else:
            Z = Z.detach().clone()
            Z[self.anchors, :] = 0.0
        return Z

    def project_onto_constraint(self, Z, inplace=True):
        # When called by our algorithms, `Z` will always be in the
        # constraint set, since we only update the non-anchors ...
        # (project_onto_tangent_space zeros out the gradient for the anchors)
        #
        # This operation is so cheap that it likely doesn't matter that we're
        # doing redundant work.
        if inplace:
            Z[self.anchors, :] = self.values
        else:
            Z = Z.detach().clone()
            Z[self.anchors, :] = self.values
        return Z


class _Standardized(Constraint):
    """Standardization constraint.

    Constrains an embedding X to be centered (the rows have mean zero)
    and to satisfy (1/n) X.T @ X == I
    """

    def name(self):
        return "standardized"

    def initialization(self, n_items, embedding_dim, device=None):
        X = torch.randn((n_items, embedding_dim), device=device)
        X -= X.mean(axis=0)
        if not isinstance(n_items, torch.Tensor):
            n_items = torch.tensor(n_items, dtype=X.dtype, device=device)
        lmbda, Q = torch.linalg.eigh(X.T @ X, UPLO="U")
        X = X @ Q @ torch.diag(lmbda ** (-0.5)) * (n_items.type(X.dtype).sqrt())
        return X

    def project_onto_tangent_space(self, X, Z, inplace=True):
        n = torch.tensor(X.shape[0], dtype=X.dtype, device=X.device)
        gtx = Z.T @ X
        if inplace:
            return Z.sub_((1.0 / n) * X @ gtx)
        else:
            return Z - (1.0 / n) * X @ gtx

    def project_onto_constraint(self, Z, inplace=True):
        return util.proj_standardized(Z, demean=True, inplace=inplace)

    def natural_length(self, n_items, embedding_dim):
        return (
            torch.tensor(2.0) * n_items * embedding_dim / (n_items - 1)
        ).sqrt()


class _Sphere(Constraint):
    def __init__(self, radius):
        self.radius = radius
        super(_Sphere, self).__init__()

    def name(self):
        return "sphere"

    def initialization(self, n_items, embedding_dim, device=None):
        X = torch.randn((n_items, embedding_dim), device=device)
        return self.radius * (X / X.norm(dim=1)[:, None])

    def project_onto_tangent_space(self, X, Z, inplace=True):
        # get the diagonal of Z @ X.T efficiently
        dual_variables = torch.bmm(
            Z.view(Z.shape[0], 1, Z.shape[1]),
            X.view(X.shape[0], X.shape[1], 1),
        ).squeeze()
        offset = (1.0 / self.radius) * dual_variables[:, None] * X
        if inplace:
            return Z.sub_(offset)
        return Z - offset

    def project_onto_constraint(self, Z, inplace=True):
        if inplace:
            Z.div_(Z.norm(dim=1)[:, None])
            Z.mul_(self.radius)
            return Z
        return self.radius * Z / Z.norm(dim=1)[:, None]


__Centered = _Centered()
__Standardized = _Standardized()


def Centered():
    """Centering constraint.

    This function returns a centering constraint, which requires the embedding
    vectors to be centered around 0.
    """
    return __Centered


def Standardized():
    """Standardization constraint.

    This function returns a standardization constraint, which constrains an
    embedding :math:`X` to be centered (the rows have mean zero) and to satisfy
    :math:`(1/n) X^T  X = I`.
    """
    return __Standardized



import argparse
import cProfile
import functools
import re
import time

import numpy as np
import torch
from pymde import constraints
from pymde import problem
from pymde import quadratic
from pymde import util
from pymde.functions import penalties


METHODS = []


def register_benchmark(n_rounds):
    def decorator(func):
        @functools.wraps(func)
        def wrapper(self):
            for i in range(n_rounds):
                np.random.seed(i)
                torch.manual_seed(i)
                func(self)

        METHODS.append(wrapper)
        return wrapper

    return decorator


def print_name(func):

    name = func.__name__[func.__name__.find("_") + 1 :]

    @functools.wraps(func)
    def wrapper(self):
        print(name, end=",")
        func(self)

    return wrapper


class Benchmark(object):
    def __init__(self, eps, max_iter, memory, regex, profile):
        self.eps = eps
        self.max_iter = max_iter
        self.memory = memory
        self.regex = regex
        self.profile = profile

    def run(self):
        print("name,device,n,m,p,seconds,iters,residual norm")
        if self.profile is not None:
            test_name = self.profile[: self.profile.rfind("_")]
            method = None
            for m in METHODS:
                if m.__name__ == test_name:
                    method = m
                    break
            if method is None:
                raise ValueError("test not found")
            device = self.profile[self.profile.rfind("_") + 1 :]
            util.set_default_device(device)
            np.random.seed(0)
            torch.manual_seed(0)
            while hasattr(method, "__wrapped__"):
                method = method.__wrapped__
            method(self)
            return

        for method in METHODS:
            name = method.__name__
            name = name[name.find("_") + 1 :]
            if re.match(self.regex, name):
                method(self)

    def _sec_iters_err(self, mde, sec):
        residual_norm = mde.solve_stats.residual_norms[-1]
        return sec, mde.solve_stats.iterations, residual_norm

    def _fit(self, mde):
        if self.profile is not None:
            pr = cProfile.Profile()
            pr.enable()
        mde.embed(
            max_iter=self.max_iter,
            eps=self.eps,
            memory_size=self.memory,
        )
        if self.profile is not None:
            pr.disable()
            pr.dump_stats(self.profile + ".prof")

    @register_benchmark(3)
    @print_name
    def test_spectral_tiny(self):
        for device in ["cpu", "cuda"]:
            n = 50
            m = 2
            p = 10 * n

            edges = util.random_edges(n, p).to(device)
            weights = np.abs(np.random.randn(p))
            f = penalties.Quadratic(
                torch.tensor(weights, dtype=torch.float)
            ).to(device)

            mde = problem.MDE(
                n_items=n,
                embedding_dim=m,
                edges=edges,
                distortion_function=f,
                constraint=constraints.Standardized(),
                device=device,
            )
            start = time.time()
            self._fit(mde)
            end = time.time()

            sec, iters, err = self._sec_iters_err(mde, end - start)
            print(
                ",".join(
                    [
                        device,
                        str(n),
                        str(m),
                        str(edges.shape[0]),
                        str(sec),
                        str(iters),
                        str(err),
                    ]
                )
            )

    @register_benchmark(3)
    @print_name
    def test_spectral_small(self):
        for device in ["cpu", "cuda"]:
            n = 500
            m = 2
            p = 10 * n

            edges = util.random_edges(n, p)
            weights = np.abs(np.random.randn(p))
            f = penalties.Quadratic(torch.tensor(weights, dtype=torch.float))

            mde = problem.MDE(
                n_items=n,
                embedding_dim=m,
                edges=edges,
                distortion_function=f,
                constraint=constraints.Standardized(),
            )
            start = time.time()
            self._fit(mde)
            end = time.time()

            sec, iters, err = self._sec_iters_err(mde, end - start)
            print(
                ",".join(
                    [
                        str(device),
                        str(n),
                        str(m),
                        str(edges.shape[0]),
                        str(sec),
                        str(iters),
                        str(err),
                    ]
                )
            )

    @register_benchmark(3)
    @print_name
    def test_spectral_medium(self):
        for device in ["cpu", "cuda"]:
            n = 5000
            m = 2
            p = 10 * n

            edges = util.random_edges(n, p).to(device)
            weights = np.abs(np.random.randn(p))
            f = penalties.Quadratic(
                torch.tensor(weights, dtype=torch.float)
            ).to(device)

            mde = problem.MDE(
                n_items=n,
                embedding_dim=m,
                edges=edges,
                distortion_function=f,
                constraint=constraints.Standardized(),
            )
            start = time.time()
            self._fit(mde)
            end = time.time()

            sec, iters, err = self._sec_iters_err(mde, end - start)
            print(
                ",".join(
                    [
                        str(device),
                        str(n),
                        str(m),
                        str(edges.shape[0]),
                        str(sec),
                        str(iters),
                        str(err),
                    ]
                )
            )

    @register_benchmark(2)
    @print_name
    def test_spectral_large(self):
        for device in ["cpu", "cuda"]:
            n = 50000
            m = 2
            p = 10 * n

            edges = util.random_edges(n, p).to(device)
            weights = np.abs(np.random.randn(p))
            f = penalties.Quadratic(
                torch.tensor(weights, dtype=torch.float)
            ).to(device)

            mde = problem.MDE(
                n_items=n,
                embedding_dim=m,
                edges=edges,
                distortion_function=f,
                constraint=constraints.Standardized(),
            )
            start = time.time()
            self._fit(mde)
            end = time.time()

            sec, iters, err = self._sec_iters_err(mde, end - start)
            print(
                ",".join(
                    [
                        str(device),
                        str(n),
                        str(m),
                        str(edges.shape[0]),
                        str(sec),
                        str(iters),
                        str(err),
                    ]
                )
            )

    @register_benchmark(1)
    @print_name
    def test_spectral_tiny_scipy(self):
        n = 50
        m = 2
        p = 10 * n

        edges = util.random_edges(n, p)
        weights = np.abs(np.random.randn(p))
        L = quadratic._laplacian(n, m, edges, weights)
        start = time.time()
        quadratic._spectral(L, m)
        end = time.time()

        print(
            ",".join(
                [
                    "cpu",
                    str(n),
                    str(m),
                    str(edges.shape[0]),
                    str(end - start),
                    str(-1),
                ]
            )
        )

    @register_benchmark(1)
    @print_name
    def test_spectral_small_scipy(self):
        n = 500
        m = 2
        p = 10 * n

        edges = util.random_edges(n, p)
        weights = np.abs(np.random.randn(p))
        L = quadratic._laplacian(n, m, edges, weights)
        start = time.time()
        quadratic._spectral(L, m)
        end = time.time()

        device = "cpu"
        print(
            ",".join(
                [
                    str(device),
                    str(n),
                    str(m),
                    str(edges.shape[0]),
                    str(end - start),
                    str(-1),
                ]
            )
        )

    @register_benchmark(1)
    @print_name
    def test_spectral_medium_scipy(self):
        n = 5000
        m = 2
        p = 10 * n

        edges = util.random_edges(n, p)
        weights = np.abs(np.random.randn(p))
        L = quadratic._laplacian(n, m, edges, weights)
        start = time.time()
        quadratic._spectral(L, m)
        end = time.time()

        device = "cpu"
        print(
            ",".join(
                [
                    str(device),
                    str(n),
                    str(m),
                    str(edges.shape[0]),
                    str(end - start),
                    str(-1),
                ]
            )
        )

    @register_benchmark(1)
    @print_name
    def test_spectral_large_scipy(self):
        n = 50000
        m = 2
        p = 10 * n

        edges = util.random_edges(n, p)
        weights = np.abs(np.random.randn(p))
        L = quadratic._laplacian(n, m, edges, weights)
        start = time.time()
        quadratic._spectral(L, m)
        end = time.time()

        device = "cpu"
        print(
            ",".join(
                [
                    str(device),
                    str(n),
                    str(m),
                    str(edges.shape[0]),
                    str(end - start),
                    str(-1),
                ]
            )
        )


def driver(eps, max_iter, memory, regex, profile):
    benchmark = Benchmark(
        eps=eps,
        max_iter=max_iter,
        memory=memory,
        regex=regex,
        profile=profile,
    )
    benchmark.run()


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Optimization benchmarks",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "--eps",
        dest="eps",
        metavar="<eps>",
        default=1e-5,
        type=float,
        help="solver tolerance",
    )
    parser.add_argument(
        "--maxiters",
        metavar="<maxiters>",
        default=300,
        type=int,
        help="max iterations",
    )
    parser.add_argument(
        "--memory",
        metavar="<memory>",
        default=10,
        type=int,
        help="LBFGS memory size",
    )
    parser.add_argument(
        "--profile",
        metavar="<testname>",
        type=str,
        help=(
            "name of test method to profile; "
            "profile will be saved to <testname>.prof; "
            "eg, test_spectral_medium_cpu or "
            "test_spectral_medium_cuda"
        ),
    )
    parser.add_argument(
        "--r",
        dest="regex",
        metavar="<regex>",
        default="",
        type=str,
        help="only run benchmarks matching regex",
    )
    args = parser.parse_args()
    driver(
        args.eps,
        args.maxiters,
        args.memory,
        args.regex,
        args.profile,
    )



import codecs
import distutils.sysconfig
import distutils.version
import os.path

from setuptools import setup, find_packages, Extension
import numpy as np


with open("README.md", "r") as fh:
    long_description = fh.read()


_graph = Extension(
    "pymde.preprocess._graph",
    sources=["pymde/preprocess/_graph.pyx"],
    # once Cython 3.0 is released, uncomment the below line
    # define_macros=[("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION")],
    extra_compile_args=["-O3"],
    include_dirs=[np.get_include()],
)


def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), "r") as fp:
        return fp.read()


def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith("__version__"):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")


setup(
    name="pymde",
    version=get_version("pymde/__init__.py"),
    description="Minimum-Distortion Embedding",
    long_description=long_description,
    long_description_content_type="text/markdown",
    setup_requires=["setuptools>=18.0", "cython"],
    install_requires=[
        "matplotlib",
        "numpy >= 1.17.5",
        "pynndescent",
        "scipy",
        "torch >= 1.7.1",
        "torchvision >= 0.8.2",
        # torchvision requires requests but does not list it as a dependency
        "requests",  
    ],
    packages=find_packages(),
    ext_modules=[_graph],
    license="Apache License, Version 2.0",
    license_files=["LICENSE"],
    url="https://github.com/cvxgrp/pymde",
    classifiers=[
        "Programming Language :: Python :: 3",
    ],
    author="Akshay Agrawal",
    author_email="akshayka@cs.stanford.edu",
)



"""Penalties: distortion functions derived from weights.

A vector distortion function  :math:`f : \\mathbf{R}^{p} \\to \\mathbf{R}^p`
derived from weights has component functions

.. math::

    f_k(d_k) = w_kp(d_k), \\quad k=1, \\ldots, p,

where :math:`w_k` is a scalar weight, :math:`p` is a penalty function, and
:math:`d_k` is an embedding distance. The penalty encourages distances to be
small when the weights are positive, and encourages them to be not small
when the weights are negative.

When an MDE problem calls a distortion function, :math:`d_k` is the Euclidean
distance between the items paired by the :math:`k`-th edge, so :math:`w_k`
should be the weight associated with the :math:`k`-th edge, and
:math:`f_k(d_k)` is the distortion associated with the edge.

Every penalty can be used with positive or negative weights. When :math:`w_k`
is positive, :math:`f_k` is attractive, meaning it encourages the embedding
distances to be small; when :math:`w_k` is negative, :math:`f_k` is repulsive,
meaning it encourages the distances to be large. Some penalties are better
suited to attracting points, while others are better suited to repelling them.

**Negative weights.**
For negative weights, it is recommended to only use one of the
following penalties:

.. code:: python3

    pymde.penalties.Log
    pymde.penalties.InvPower
    pymde.penalties.LogRatio

These penalties go to negative infinity as the input approaches zero,
and to zero as the input approaches infinity. With a negative weight,
that means the distortion function goes to infinity at 0, and to 0 at infinity.

Using other penalties with negative weights is possible, but it can lead to
pathological MDE problems if you are not careful.

**Positive weights.**
Penalties that work well in attracting points are those that are :math:`0`
when the distance is :math:`0`, grows when the distance is larger than
:math:`0`. All the penalties in this module, other than the ones listed above
(and the function described below), can be safely used with attractive
penalties. Some examples inlcude:

.. code:: python3

    pymde.penalties.Log1p
    pymde.penalties.Linear
    pymde.penalties.Quadratic
    pymde.penalties.Cubic
    pymde.penalties.Huber

**Combining penalties.**
The ``PushAndPull`` function can be used to combine two penalties, an attractive
penalty for use with positive weights, and a repulsive penalty for use with
negative weights. This leads to a distortion function of the form

.. math::

    f_k(d) = \\begin{cases}
        w_k p_{\\text{attractive}}(d_k) & w_k > 0 \\\\
        w_k p_{\\text{repulsive}}(d_k) & w_k < 0 \\\\
    \\end{cases}.

For example:

.. code:: python3

    weights = torch.tensor([1., 1., -1., 1., -1.])
    attractive_penalty = pymde.penalties.Log1p
    repulsive_penalty = pymde.penalties.Log

    distortion_function = pymde.PushAndPull(
        weights,
        attractive_penalty,
        repulsive_penalty)

**Example.**
Distortion functions are created in a vectorized or elementwise fashion. The
constructor takes a sequence (torch.Tensor) of weights, returning a callable
object. The object takes a sequence of distances of the same length as the
weights, and returns a sequence of distortions, one for each distance.

For example:

.. code:: python3

    weights = torch.tensor([1., 2., 3.])
    f = pymde.penalties.Quadratic(weights)

    distances = torch.tensor([2., 1., 4.])
    distortions = f(distances)
    # the distortions are 1 * 2**2 == 4, 2 * 1**2 == 2, 3 * 4**2 = 48
    print(distortions)

prints

.. code:: python3

    torch.tensor([4., 2., 48.])
"""
from pymde import util
from pymde.functions.function import Function
import torch


class Linear(Function):
    """:math:`p(d) = d`"""

    def __init__(self, weights):
        super(Linear, self).__init__()
        self.weights = util.to_tensor(weights)

    def forward(self, distances):
        return self.weights * distances


class Quadratic(Function):
    """:math:`p(d) = d^2`"""

    def __init__(self, weights):
        super(Quadratic, self).__init__()
        self.weights = util.to_tensor(weights)

    def forward(self, distances):
        return self.weights * distances.pow(2)


class _DeadzoneQuadratic(Function):
    def __init__(self, weights, threshold):
        super(_DeadzoneQuadratic, self).__init__()
        self.weights = util.to_tensor(weights)
        self.threshold = threshold

    def forward(self, distances):
        output = torch.zeros(
            distances.shape, dtype=distances.dtype, device=distances.device
        )
        lt_thresh = distances < self.threshold
        gt_thresh = ~lt_thresh
        output[lt_thresh] = 0.0
        output[gt_thresh] = distances[gt_thresh].pow(2)
        return self.weights * output


class _ClippedQuadratic(Function):
    def __init__(self, weights, threshold):
        super(_ClippedQuadratic, self).__init__()
        self.weights = util.to_tensor(weights)
        self.threshold = threshold

    def forward(self, distances):
        return self.weights * torch.min(
            distances**2, (self.threshold + 1) ** 2
        )


class Cubic(Function):
    """:math:`p(d) = d^3`"""

    def __init__(self, weights):
        super(Cubic, self).__init__()
        self.weights = util.to_tensor(weights)

    def forward(self, distances):
        return self.weights * distances.pow(3)


class _DeadzoneCubic(Function):
    def __init__(self, weights, threshold):
        super(_DeadzoneCubic, self).__init__()
        self.weights = util.to_tensor(weights)
        self.threshold = threshold

    def forward(self, distances):
        output = torch.zeros(
            distances.shape, dtype=distances.dtype, device=distances.device
        )
        lt_thresh = distances < self.threshold
        gt_thresh = ~lt_thresh
        output[lt_thresh] = 0.0
        output[gt_thresh] = distances[gt_thresh].pow(3)
        return self.weights * output


class Power(Function):
    """:math:`p(d) = d^\\text{exponent}`"""

    def __init__(self, weights, exponent):
        super(Power, self).__init__()
        self.weights = util.to_tensor(weights)
        if not isinstance(exponent, torch.Tensor):
            exponent = torch.tensor(exponent, device=self.weights.device)
        self.exponent = exponent

    def forward(self, distances):
        return self.weights * distances.pow(self.exponent)


class Huber(Function):
    """
    .. math::

        p(d) = \\begin{cases}
            0.5 \\cdot d^2 & d < \\text{threshold} \\\\
            \\text{threshold}(d - 0.5 \\cdot \\text{threshold})
            & d \\geq \\text{threshold}
        \\end{cases}

    """

    def __init__(self, weights, threshold=0.5):
        if threshold < 0:
            raise ValueError(
                "Threshold must be nonnegative, received ", threshold
            )
        super(Huber, self).__init__()
        self.weights = util.to_tensor(weights)
        self.threshold = threshold

    def forward(self, distances):
        output = torch.zeros(
            distances.shape, dtype=distances.dtype, device=distances.device
        )
        lt_idx = distances < self.threshold
        if self.weights.nelement() == 1:
            lt_weights = self.weights
            gt_weights = self.weights
        else:
            lt_weights = self.weights[lt_idx]
            gt_weights = self.weights[~lt_idx]
        output[lt_idx] = lt_weights * 0.5 * distances[lt_idx].pow(2)
        output[~lt_idx] = (
            gt_weights
            * self.threshold
            * (distances[~lt_idx] - 0.5 * self.threshold)
        )
        return output


class Logistic(Function):
    """:math:`p(d) = \\log(1 + e^{\\alpha(d - \\text{threshold})})`"""

    def __init__(self, weights, threshold=0.0, alpha=3.0):
        if threshold < 0:
            raise ValueError(
                "Threshold must be nonnegative, received ", threshold
            )
        super(Logistic, self).__init__()
        self.weights = util.to_tensor(weights)
        self.threshold = threshold
        self.alpha = alpha

    def forward(self, distances):
        zeros = torch.zeros(
            distances.shape, dtype=distances.dtype, device=distances.device
        )
        stacked = torch.stack(
            (zeros, self.alpha * (distances - self.threshold))
        ).T
        return self.weights * torch.logsumexp(stacked, dim=1)


class Sigmoid(Function):
    def __init__(self, weights, threshold, alpha=1.0):
        if threshold < 0:
            raise ValueError(
                "Threshold must be nonnegative, received ", threshold
            )
        super(Sigmoid, self).__init__()
        self.weights = util.to_tensor(weights)
        self.threshold = threshold
        self.alpha = alpha

    def forward(self, distances):
        return self.weights * torch.sigmoid(
            self.alpha * (distances - self.threshold)
        )


class Hinge(Function):
    def __init__(self, weights, threshold, sigma=None):
        if threshold < 0:
            raise ValueError(
                "Threshold must be nonnegative, received ", threshold
            )
        if sigma is None:
            sigma = threshold / 2
        super(Hinge, self).__init__()
        self.weights = util.to_tensor(weights)
        self.threshold = threshold
        self.sigma = sigma

    def forward(self, distances):
        return torch.max(
            torch.tensor(0.0, device=distances.device, dtype=distances.dtype),
            self.weights
            * (
                distances
                - (self.threshold - torch.sign(self.weights) * (self.sigma))
            ),
        )


class Log1p(Function):
    """:math:`p(d) = \log(1 + d^{\\text{exponent}})`"""

    def __init__(self, weights, exponent=1.5):
        super(Log1p, self).__init__()
        self.weights = util.to_tensor(weights)
        if not isinstance(exponent, torch.Tensor):
            exponent = torch.tensor(exponent, device=self.weights.device)
        self.exponent = exponent

    def forward(self, distances):
        return self.weights * torch.log1p(distances.pow(self.exponent))


class Log(Function):
    """:math:`p(d) = \log(1 - \\exp(-d^\\text{exponent}))`"""

    def __init__(self, weights, exponent=1.0):
        super(Log, self).__init__()
        self.weights = util.to_tensor(weights)
        if not isinstance(exponent, torch.Tensor):
            exponent = torch.tensor(exponent, device=weights.device)
        self.exponent = exponent

    def forward(self, distances):
        return self.weights * torch.log(
            -torch.expm1(-(distances**self.exponent))
        )


class InvPower(Function):
    """:math:`p(d) = 1/d^\\text{exponent}`"""

    def __init__(self, weights, exponent=1):
        if not (weights <= 0).all():
            raise ValueError("Weights must be negative.")
        super(InvPower, self).__init__()
        self.weights = util.to_tensor(weights)
        if not isinstance(exponent, torch.Tensor):
            exponent = torch.tensor(exponent)
        self.exponent = exponent

    def forward(self, distances):
        return self.weights.abs() * 1 / (distances**self.exponent)


class LogRatio(Function):
    """:math:`p(d) = \\log\\left(\\frac{d^\\text{exponent}}{1 + d^{\\text{exponent}}}\\right)`"""  # noqa: E501

    def __init__(self, weights, exponent=2):
        super(LogRatio, self).__init__()
        self.weights = util.to_tensor(weights)
        if not isinstance(exponent, torch.Tensor):
            exponent = torch.tensor(exponent)
        self.exponent = exponent

    def forward(self, distances):
        return self.weights * torch.log(
            distances**self.exponent / (1 + distances**self.exponent)
        )


class PushAndPull(Function):
    """Combine an attractive and repulsive penalty.

        .. math::

            f_k(d) = \\begin{cases}
                w_k p_{\\text{attractive}}(d_k) & w_k > 0 \\\\
                w_k p_{\\text{repulsive}}(d_k) & w_k < 0 \\\\
            \\end{cases}
    """

    def __init__(
        self, weights, attractive_penalty=Log1p, repulsive_penalty=LogRatio
    ):
        super(PushAndPull, self).__init__()
        self.weights = util.to_tensor(weights)
        if weights.nelement() == 1:
            raise ValueError("`PushAndPull` requires at least two weights.")
        self.pos_idx = weights >= 0
        self.attractive_penalty = attractive_penalty(weights[self.pos_idx])
        self.repulsive_penalty = repulsive_penalty(weights[~self.pos_idx])

    def forward(self, distances):
        output = torch.empty(
            distances.shape, dtype=distances.dtype, device=distances.device
        )
        output[self.pos_idx] = self.attractive_penalty(distances[self.pos_idx])
        output[~self.pos_idx] = self.repulsive_penalty(distances[~self.pos_idx])
        return output



import sys

import numpy as np
import scipy.sparse as sp
import torch


from pymde import constraints
from pymde import problem
from pymde import util
from pymde.functions import penalties, losses
from pymde.preprocess import _graph


__this_module = sys.modules[__name__]


LOGGER = problem.LOGGER


def _to_edges(graph):
    if not sp.issparse(graph):
        n = graph.shape[0]
        edges = torch.triu_indices(n, n, 1).T
        distances = graph[edges]
        return edges, distances

    # discard the lower triangular part, which is redundant
    graph = sp.triu(graph, format="coo")
    edges = torch.stack(
        [
            torch.tensor(graph.row, dtype=torch.long),
            torch.tensor(graph.col, dtype=torch.long),
        ],
        dim=1,
    )
    data = torch.tensor(graph.data, dtype=torch.float)
    return edges, data


def _validate_adjacency_matrix(matrix):
    diagonal_nonzero = matrix.diagonal() > 0
    if diagonal_nonzero.any():
        raise ValueError(
            "Adjacency matrices must not contain self edges; "
            "the following nodes were found to have self edges: ",
            np.argwhere(diagonal_nonzero).flatten(),
        )


def _to_graph(edges, distances=None, n_items=None):
    "Distances for repeated edges are summed."
    if isinstance(edges, torch.Tensor):
        edges = edges.cpu().numpy()

    if distances is None:
        distances = np.ones(edges.shape[0], dtype=np.float32)
    elif isinstance(distances, torch.Tensor):
        distances = distances.cpu().float().numpy()

    flip_idx = edges[:, 0] > edges[:, 1]
    edges[flip_idx] = np.stack(
        [edges[flip_idx][:, 1], edges[flip_idx][:, 0]], axis=1
    )

    if n_items is None:
        n_items = edges.max() + 1
    rows = edges[:, 0]
    cols = edges[:, 1]
    graph = sp.coo_matrix((distances, (rows, cols)), shape=(n_items, n_items))
    graph = graph + graph.T
    return Graph(graph.tocsr())


class Graph(object):
    """A weighted graph.

    This class represents a weighted graph. It is backed by a scipy.sparse
    adjacency matrix, and can be constructed from either a dense
    distance matrix, a sparse adjacency matrix, or torch.Tensor's
    of edges and weights (using the `from_edges` static method).

    It is an error for the graph to contain self-edges (i.e., non-zero
    values on the diagonal).

    This class is accepted as an argument to various preprocessing functions.
    """

    def __init__(self, adjacency_matrix):
        if isinstance(adjacency_matrix, np.ndarray):
            adjacency_matrix = sp.csr_matrix(adjacency_matrix)
        elif isinstance(adjacency_matrix, torch.Tensor):
            adjacency_matrix = sp.csr_matrix(adjacency_matrix.cpu().numpy())
        elif sp.issparse(adjacency_matrix) and not isinstance(
            adjacency_matrix, sp.csr_matrix
        ):
            LOGGER.warning(
                "The adjacency matrix for a Graph object should be "
                "a CSR matrix; converting your matrix to a CSR "
                "matrix, which may be costly ..."
            )
            adjacency_matrix = adjacency_matrix.tocsr()

        unreachable = adjacency_matrix.data == np.inf
        adjacency_matrix.data[unreachable] = 0
        adjacency_matrix.eliminate_zeros()
        _validate_adjacency_matrix(adjacency_matrix)

        self._adjacency_matrix = adjacency_matrix
        self._edges = None
        self._distances = None

    @staticmethod
    def from_edges(edges, weights=None, n_items=None):
        """Construct a graph from edges and weights.

        Arguments
        ---------
        edges: torch.Tensor
            Tensor of edges, of shape ``(n_edges, 2)``, with each
            edge represented by a row, i.e. by two integers.
        weights: torch.Tensor, optional
            Tensor of weights, of shape ``(n_edges,)``, with each weight
            a float.
        n_items: int, optional
            The number of items in the graph; if ``None``, this is
            taken to be the maximum value in ``edges`` plus 1.

        Returns
        -------
        pymde.Graph
            A Graph object representing ``edges`` and ``weights``.
        """
        return _to_graph(edges, weights, n_items)

    @property
    def adjacency_matrix(self):
        """The scipy.sparse adjacency matrix backing the graph."""
        return self._adjacency_matrix

    @property
    def A(self):
        return self.adjacency_matrix

    @A.setter
    def A(self, value):
        self.adjacency_matrix = value

    @property
    def edges(self):
        """A torch.Tensor of the edges in the graph."""
        if self._edges is None:
            edges, distances = _to_edges(self.A)
            self._edges = edges
            self._distances = distances
        return self._edges

    @property
    def distances(self):
        """The distances associated with each edge."""
        if self._distances is None:
            edges, distances = _to_edges(self.A)
            self._edges = edges
            self._distances = distances
        return self._distances

    @property
    def weights(self):
        return self.distances

    @property
    def n_items(self):
        """The number of items in the graph."""
        return self.A.shape[0]

    @property
    def n_edges(self):
        """The number of edges in the graph."""
        return self.edges.shape[0]

    @property
    def n_all_edges(self):
        """n_items choose 2."""
        return self.n_items * (self.n_items - 1) // 2

    def neighbors(self, node: int) -> np.ndarray:
        """The indices of the neighbors of ``node``."""
        return self.A.indices[self.A.indptr[node] : self.A.indptr[node + 1]]

    def neighbor_distances(self, node) -> np.ndarray:
        """The distances associated with the edges connected to ``node``."""
        return self.A.data[self.A.indptr[node] : self.A.indptr[node + 1]]

    def __getitem__(self, key):
        return self.A[key]

    def __setitem__(self, key, value):
        raise AttributeError("Graph objects are immutable.")

    def draw(
        self, embedding_dim=2, standardized=False, device="cpu", verbose=False
    ):
        """Draw a graph in the Cartesian plane.

        This method does some basic preprocessing, constructs an MDE problem
        that is often suitable for drawing graphs, and computes/returns an
        embedding by approximately solving the MDE problem.

        Arguments
        ---------
        embedding_dim: int
            The number of dimemsions, 1, 2, or 3.
        standardized: bool
            Whether to impose a standardization constraint.
        device: str
            Device on which to compute/store embedding, 'cpu' or 'cuda'.
        verbose: bool
            Whether to print verbose output.

        Returns
        -------
        torch.Tensor
            The embedding, of shape ``(n_items, embedding_dim)``
        """
        if (self.distances < 0).any():
            raise ValueError(
                "Graphs with negative edge weights cannot be drawn."
            )

        if self.n_edges < self.n_all_edges:
            # copmute shortest paths, but no more than 1e7 of them ...
            retain_fraction = min(1.0, 1e7 / self.n_all_edges)
            distance_graph = shortest_paths(
                self, retain_fraction=retain_fraction, verbose=verbose
            )
        else:
            distance_graph = self

        if not standardized:
            constraint = constraints.Centered()
            f = losses.WeightedQuadratic(distance_graph.distances)
        else:
            constraint = constraints.Standardized()
            # TODO(akshayka) better weights
            f = penalties.Cubic(1 / distance_graph.distances)
        mde = problem.MDE(
            n_items=self.n_items,
            embedding_dim=embedding_dim,
            edges=distance_graph.edges,
            distortion_function=f,
            constraint=constraint,
            device=device,
        )
        X = mde.embed(verbose=verbose)
        mde.plot(edges=self.edges)
        return X


def scale(graph, natural_length):
    """Scale graph weights to have RMS equal to `natural_length`

    Returns a new graph, leaving the old graph unmodified.

    Arguments
    ---------
    graph: pymde.Graph
        The graph whose distances to scale.
    natural_length: float
        Target RMS value.

    Returns
    -------
    pymde.Graph
        A new graph with scaled distances.
    """
    rms = graph.distances.pow(2).mean().sqrt()
    alpha = natural_length / rms
    distances = alpha * graph.distances
    return _to_graph(graph.edges, distances.cpu().numpy())


def _sparsify(graph, retain_fraction):
    raise NotImplementedError


def breadth_first_order(
    csgraph, i_start, directed=True, return_predecessors=True
):
    N = csgraph.shape[0]
    node_list = np.empty(N, dtype=np.int32)
    lengths = np.empty(N, dtype=np.int32)
    predecessors = np.empty(N, dtype=np.int32)
    node_list.fill(-9999)
    lengths.fill(-9999)
    predecessors.fill(-9999)

    if directed:
        _graph._breadth_first_directed(
            i_start,
            csgraph.indices,
            csgraph.indptr,
            node_list,
            lengths,
            predecessors,
        )
    lengths = lengths.astype(np.float32)
    lengths[lengths < 0] = np.inf
    return lengths, predecessors


def _shortest_paths(shape, node, max_length, unweighted, sample_prob):
    data = __this_module.__data
    indptr = __this_module.__indptr
    indices = __this_module.__indices

    data, indptr, indices = map(np.ctypeslib.as_array, [data, indptr, indices])
    A = sp.csr_matrix((data, indices, indptr), shape=shape)
    if unweighted:
        # fast path
        distances, _ = breadth_first_order(
            A,
            node,
        )
        distances[distances > max_length] = np.inf
    else:
        distances = sp.csgraph.dijkstra(A, indices=node, limit=max_length)

    indices = np.argwhere((distances != np.inf) * (distances > 0)).ravel()
    indices = indices[indices > node]
    if sample_prob < 1.0:
        sample_mask = util.np_rng().uniform(size=indices.size) <= sample_prob
        indices = indices[sample_mask]

    distances = distances[indices]
    return distances, indices


def __init_process(data, indptr, indices):
    global __this_module
    __this_module.__data = data
    __this_module.__indptr = indptr
    __this_module.__indices = indices


def shortest_paths(
    graph, max_length=None, retain_fraction=1.0, n_workers=None, verbose=False
):
    """Compute shortest-path distances.

    This function computes the shortest-path distances on a graph.
    It returns a new graph, with an edge for each distance that was
    computed (each edge is weighted by the shortest-path distance between
    the two nodes in the original graph). The new graph can be interpreted
    as a "filled-in" version of the input graph.

    The ``max_length`` and ``retain_fraction`` keyword arguments can be used
    to compute only a subset of the distances. This can be useful
    if the graph has a very large number of nodes, in which storing all graph
    distances may be intractable.

    If the graph is sufficiently large, multiple cores will be used to
    accelerate the computation.

    Arguments
    ---------
    graph: pymde.Graph or scipy.sparse matrix
        Graph instance, or an adjacency matrix
    max_length: float, optional
        The maximum-length path to compute; paths longer than max_length
        are not computed/explored.
    retain_fraction: float
        A float between 0 and 1, specifying the fraction of all (n choose 2)
        to compute. For example, if ``retain_fraction`` is 0.1, only 10
        percent of the edges will be stored.
    n_workers: int, optional
        The number of processes to use. Defaults to the number of available
        cores.
    verbose: bool
        Whether to print verbose output.

    Returns
    -------
    pymde.Graph
        A new graph, with an edge (and weight) for each shortest-path
        distance that was computed/stored.
    """
    if sp.issparse(graph):
        graph = Graph(graph)
    elif not isinstance(graph, Graph):
        raise ValueError(
            "`graph` must be a pymde.Graph instance or "
            "scipy.sparse adjacency matrix."
        )

    if verbose:
        LOGGER.info(
            f"Computing shortest path distances (retaining "
            f"{100*retain_fraction:.2f} percent with "
            f"max_distance={max_length}) ..."
        )

    if n_workers is None and max_length is None and retain_fraction == 1.0:
        # obtain a dense distance matrix
        distance_matrix = sp.csgraph.shortest_path(graph.A, directed=False)
        return Graph(sp.csr_matrix(distance_matrix))
    elif max_length is None:
        max_length = np.inf
    unweighted = (graph.distances == 1.0).all()
    if verbose and not unweighted:
        LOGGER.info("Graph is weighted ... using slow path.")

    # call dijkstra n times and compress the output in each iteration
    # (Calling just once, with indices = [1, ..., n], would return a dense
    # n times n matrix.)
    import multiprocessing
    from multiprocessing import sharedctypes

    def to_shared_memory(array):
        array = np.ctypeslib.as_ctypes(array)
        return sharedctypes.RawArray(array._type_, array)

    global __this_module
    data = __this_module.__data = to_shared_memory(graph.A.data)
    indptr = __this_module.__indptr = to_shared_memory(graph.A.indptr)
    indices = __this_module.__indices = to_shared_memory(graph.A.indices)

    nodes = np.arange(graph.n_items)
    print_every = max(1, nodes.shape[0] // 10)
    n_workers = multiprocessing.cpu_count() if n_workers is None else n_workers
    with multiprocessing.Pool(
        n_workers, initializer=__init_process, initargs=(data, indptr, indices)
    ) as pool:
        try:
            async_ret = [
                pool.apply_async(
                    _shortest_paths,
                    (
                        graph.A.shape,
                        node,
                        max_length,
                        unweighted,
                        retain_fraction,
                    ),
                )
                for node in nodes
            ]
            values = []
            for idx, ret in enumerate(async_ret):
                data, columns = ret.get()
                values.append((data, columns))
                if verbose and idx % print_every == 0:
                    LOGGER.info(f"processed node {idx + 1}/{nodes.shape[0]}")
        except:  # noqa: E722
            del __this_module.__data
            del __this_module.__indptr
            del __this_module.__indices
            raise

    del __this_module.__data
    del __this_module.__indptr
    del __this_module.__indices

    data, columns = zip(*values)
    graph.A.eliminate_zeros()

    indptr = [0]
    for row in range(graph.n_items):
        indptr.append(indptr[-1] + columns[row].size)
    data = np.concatenate(data)
    indices = np.concatenate(columns)
    adj_matrix = sp.csr_matrix((data, indices, indptr), shape=graph.A.shape)

    adj_matrix = adj_matrix + adj_matrix.T.tocsr()
    return Graph(adj_matrix)


def _minimum_spanning_tree(graph):
    return Graph(sp.csgraph.minimum_spanning_tree(graph.A))


def _thresholded_neighborhood_graph(graph, threshold, max_length):
    graph = shortest_paths(graph, max_length)

    data = []
    indices = []
    indptr = [0]
    n = graph.n_items
    for node in range(n):
        distances = graph.neighbor_distances(node)
        columns = graph.neighbors(node)
        thresholded_indices = distances[distances <= threshold]

        data.append(distances[thresholded_indices])
        indices.append(columns[thresholded_indices])
        indptr.append(indptr[-1] + thresholded_indices.size)

    data = np.concatenate(data)
    indices = np.concatenate(indices)
    indptr = np.array(indptr)
    return Graph(sp.csr_matrix((data, indices, indptr), shape=graph.A.shape))


def k_nearest_neighbors(
    graph,
    k,
    graph_distances=False,
    max_distance=None,
    verbose=False,
):
    """Compute k-nearest neighbors for each node in graph.

    This function computes a k-nearest neighbor graph.

    By default, the input graph is interpreted as a distance matrix, with
    ``graph.adjacency_matrix[i, j]`` giving the distance between i and j. If
    `graph_distances` is True, then the shortest-path metric is used to
    compute neighbors.

    In the returned graph, if i is a neighbor of j and j a neighbor of i, then
    the weight w_{ij} will be +2; if only one is a neighbor of the other, then
    w_{ij} will be +1; if neither are neighbors of each other, then (i, j) will
    not be in the graph.

    Arguments
    ---------
    graph: pymde.Graph
        The graph, representing a distance metric on the items.
    k: int
        The number of nearest neighbors per item.
    graph_distances: bool
        If ``True``, computes shortest-path distances on the graph; otherwise,
        interprets the graph as a distance matrix.
    max_distance: float (optional)
        If not ``None``, neighborhoods are restricted to have a radius
        no greater than max_distance.
    verbose: bool
        If ``True``, print verbose output.

    Returns
    -------
    pymde.Graph
        The k-nearest neighbor graph.
    """

    if sp.issparse(graph):
        graph = Graph(graph)
    elif not isinstance(graph, Graph):
        raise ValueError(
            "`graph` must be a pymde.Graph instance or "
            "scipy.sparse adjacency matrix."
        )

    if graph.n_edges == graph.n_all_edges:
        # the graph is already full; no need to compute shortest paths.
        graph_distances = False

    if graph_distances:
        graph = shortest_paths(graph, max_length=max_distance, verbose=verbose)

    if max_distance is None:
        max_distance = np.inf

    coo_data = []
    coo_rows = []
    coo_cols = []
    for node in range(graph.n_items):
        dist = graph.neighbor_distances(node)
        cols = graph.neighbors(node)
        if cols.size == 0:
            continue

        dist[dist > max_distance] = np.inf
        knn_indices = np.argsort(dist)[:k]

        coo_rows.extend([node for _ in range(knn_indices.size)])
        coo_cols.extend([c for c in cols[knn_indices]])
        coo_data.extend([d for d in dist[knn_indices]])
    coo_graph = sp.coo_matrix(
        (coo_data, (coo_rows, coo_cols)), shape=graph.A.shape
    )

    # weights will be 0, 1, or 2
    coo_graph.data = np.ones(coo_graph.data.shape)
    coo_graph = coo_graph + coo_graph.T
    csr_graph = coo_graph.tocsr()
    return Graph(csr_graph)


# TODO(akshayka) including both threshold and max_length is confusing. fix
# this api
def _neighborhood_graph(
    graph, n_neighbors=None, threshold=None, max_length=None
):
    if n_neighbors is None and threshold is None:
        raise ValueError("either n_neighbors or threshold must be non-None")
    if n_neighbors is not None and threshold is not None:
        raise ValueError(
            "only one of n_neighbors and threshold can be non-none"
        )

    if n_neighbors is not None:
        return k_nearest_neighbors(graph, n_neighbors, max_length)
    return _thresholded_neighborhood_graph(graph, threshold, max_length)
